"""Methods for estimating statistics"""

import numpy as np
import torch as T

from pearll.common.enumerations import TrajectoryType
from pearll.common.type_aliases import Tensor
from pearll.common.utils import to_numpy, to_torch


def sample_forward_kl_divergence(
    target_dist_prob: Tensor, sample_dist_prob: Tensor, dtype: str = "torch"
) -> T.Tensor:
    """
    Sample forward KL Divergence element-wise
    https://towardsdatascience.com/approximating-kl-divergence-4151c8c85ddd

    :param target_dist_prob: probabilities from target distribution, p(x)
    :param sample_dist_prob: probabilites from sample distribution, q(x)
    :return: element-wise kl approximation
    """
    ratio = target_dist_prob / sample_dist_prob
    if TrajectoryType(dtype.lower()) == TrajectoryType.TORCH:
        ratio = to_torch(ratio)
        return ratio * T.log(ratio) - (ratio - 1)
    elif TrajectoryType(dtype.lower()) == TrajectoryType.NUMPY:
        ratio = to_numpy(ratio)
        return ratio * np.log(ratio) - (ratio - 1)
    else:
        raise ValueError(
            f"`dtype` flag should be 'torch' or 'numpy', but received {dtype}"
        )


def sample_reverse_kl_divergence(
    target_dist_prob: Tensor, sample_dist_prob: Tensor, dtype: str = "torch"
) -> T.Tensor:
    """
    Sample reverse KL Divergence element-wise
    https://towardsdatascience.com/approximating-kl-divergence-4151c8c85ddd

    :param target_dist_prob: probabilities from target distribution, p(x)
    :param sample_dist_prob: probabilites from sample distribution, q(x)
    :return: element-wise kl approximation
    """
    ratio = target_dist_prob / sample_dist_prob
    if TrajectoryType(dtype.lower()) == TrajectoryType.TORCH:
        ratio = to_torch(ratio)
        return (ratio - 1) - T.log(ratio)
    elif TrajectoryType(dtype.lower()) == TrajectoryType.NUMPY:
        ratio = to_numpy(ratio)
        return (ratio - 1) - np.log(ratio)
    else:
        raise ValueError(
            f"`dtype` flag should be 'torch' or 'numpy', but received {dtype}"
        )



from pearll.agents.a2c import A2C
from pearll.agents.adames import AdamES
from pearll.agents.base_agents import BaseAgent
from pearll.agents.cem_rl import CEM_RL
from pearll.agents.ddpg import DDPG
from pearll.agents.dqn import DQN
from pearll.agents.dyna import DynaQ
from pearll.agents.es import ES
from pearll.agents.ga import GA
from pearll.agents.ppo import PPO

__all__ = [
    "A2C",
    "BaseAgent",
    "CEM_RL",
    "DDPG",
    "DQN",
    "ES",
    "GA",
    "PPO",
    "AdamES",
    "DynaQ",
]



"""Methods for estimating the advantage function"""

from typing import Tuple

import torch as T

from pearll.common.type_aliases import Tensor


def generalized_advantage_estimate(
    rewards: Tensor,
    old_values: Tensor,
    new_values: Tensor,
    dones: Tensor,
    gamma: float = 0.99,
    gae_lambda: float = 0.95,
) -> Tuple[Tensor, Tensor]:
    """
    Generalized advantage estimate of a trajectory: https://towardsdatascience.com/generalized-advantage-estimate-maths-and-code-b5d5bd3ce737

    :param old_values: value function result with old_state input
    :param new_values: value function result with new_state input
    :param rewards: agent reward of taking actions in the environment
    :param dones: the done values of each step of the trajectory, indicates whether to bootstrap
    :param gamma: trajectory discount
    :param gae_lambda: exponential mean discount
    :return advantage: the advantage of taking actions in the environment
    :return value: the value of taking actions in the environment
    """
    dones = 1 - dones
    batch_size = rewards.shape[0]

    advantage = (
        T.zeros(batch_size + 1) if rewards.ndim == 1 else T.zeros((batch_size + 1, 1))
    )

    for t in reversed(range(batch_size)):
        delta = rewards[t] + (gamma * new_values[t] * dones[t]) - old_values[t]
        advantage[t] = delta + (gamma * gae_lambda * advantage[t + 1] * dones[t])

    value_target = advantage[:batch_size] + old_values

    assert (
        advantage[:batch_size].shape == old_values.shape
    ), f"The advantage's shape should be {old_values.shape}, instead it is {advantage[:batch_size].shape}."
    assert (
        value_target.shape == old_values.shape
    ), f"The returns' shape should be {old_values.shape}, instead it is {value_target.shape}."

    return advantage[:batch_size], value_target



import copy
from typing import List, Optional, Union

import numpy as np
import torch as T
from gym.spaces import Box, Discrete, MultiDiscrete, Space

from pearll.common.enumerations import Distribution
from pearll.common.type_aliases import Tensor
from pearll.common.utils import (
    get_device,
    get_space_range,
    get_space_shape,
    to_numpy,
    to_torch,
)
from pearll.models.encoders import IdentityEncoder, MLPEncoder
from pearll.models.heads import (
    BaseActorHead,
    BaseCriticHead,
    CategoricalHead,
    DiagGaussianHead,
    DummyHead,
)
from pearll.settings import PopulationSettings


class Model(T.nn.Module):
    def __init__(
        self,
        encoder: T.nn.Module,
        torso: T.nn.Module,
        head: T.nn.Module,
    ) -> None:
        super().__init__()
        self.encoder = encoder
        self.torso = torso
        self.head = head

    def forward(
        self, observations: Tensor, actions: Optional[Tensor] = None
    ) -> T.Tensor:
        out = self.encoder(observations, actions)
        out = self.torso(out)
        return self.head(out)


class Critic(T.nn.Module):
    """
    The critic network which approximates the Q or Value functions.

    :param encoder: the encoder network
    :param torso: the torso network
    :param head: the head network
    :param create_target: whether to create a target network
    :param polyak_coeff: the polyak coefficient for the target network
    :param device: the device to use
    """

    def __init__(
        self,
        encoder: T.nn.Module,
        torso: T.nn.Module,
        head: BaseCriticHead,
        create_target: bool = False,
        polyak_coeff: float = 0.995,
        device: Union[T.device, str] = "auto",
    ):
        super().__init__()
        self.polyak_coeff = polyak_coeff
        self.device = get_device(device)
        self.model = Model(encoder, torso, head).to(self.device)
        self.state_info = {}
        self.make_state_info()
        self.state = np.concatenate(
            [to_numpy(d.flatten()) for d in self.model.state_dict().values()]
        )
        self.space = Box(low=-1e6, high=1e6, shape=self.state.shape)
        self.space_shape = get_space_shape(self.space)
        self.space_range = get_space_range(self.space)

        # Create the target network
        self.target = None
        if create_target:
            self.target = copy.deepcopy(self.model)
            for target in self.target.parameters():
                target.requires_grad = False
            self.assign_targets()

    def make_state_info(self) -> None:
        """Make the state info dictionary"""
        start_idx = 0
        for k, v in self.model.state_dict().items():
            self.state_info[k] = (v.shape, (start_idx, start_idx + v.numel()))
            start_idx += v.numel()

    def set_state(self, state: np.ndarray) -> "Actor":
        """
        Set the state of the individual

        :param state: the state to set
        :return: self
        """
        self.state = state
        state = to_torch(state, device=self.device)
        state_dict = {
            k: state[v[1][0] : v[1][1]].reshape(v[0])
            for k, v in zip(self.state_info.keys(), self.state_info.values())
        }
        self.model.load_state_dict(state_dict)
        return self

    def numpy(self) -> np.ndarray:
        """Get the numpy representation of the individual"""
        return self.state

    def assign_targets(self) -> None:
        """Assign the target parameters"""
        self.target.load_state_dict(self.model.state_dict())

    def update_targets(self) -> None:
        """Update the target parameters"""
        # target_params = polyak_coeff * target_params + (1 - polyak_coeff) * online_params
        for online, target in zip(self.model.parameters(), self.target.parameters()):
            target.data.mul_(self.polyak_coeff)
            target.data.add_((1 - self.polyak_coeff) * online.data)

    def forward_target(
        self, observations: Tensor, actions: Optional[Tensor] = None
    ) -> T.Tensor:
        """
        Run a forward pass of the target network

        :param observations: the observations
        :param actions: the optional actions
        """
        return self.target(observations, actions)

    def forward(
        self, observations: Tensor, actions: Optional[Tensor] = None
    ) -> T.Tensor:
        return self.model(observations, actions)


class Actor(Critic):
    """
    The actor network which determines what actions to take in the environment.

    :param encoder: the encoder network
    :param torso: the torso network
    :param head: the head network
    :param create_target: whether to create a target network
    :param polyak_coeff: the polyak coefficient for the target network
    :param device: the device to use
    """

    def __init__(
        self,
        encoder: T.nn.Module,
        torso: T.nn.Module,
        head: BaseActorHead,
        create_target: bool = False,
        polyak_coeff: float = 0.995,
        device: Union[T.device, str] = "auto",
    ):
        super().__init__(
            encoder=encoder,
            torso=torso,
            head=head,
            create_target=create_target,
            polyak_coeff=polyak_coeff,
            device=device,
        )

    def action_distribution(
        self, observations: Tensor
    ) -> Optional[T.distributions.Distribution]:
        """Get the action distribution, returns None if deterministic"""
        latent_out = self.model.torso(self.model.encoder(observations))
        return self.model.head.action_distribution(latent_out)


class EpsilonGreedyActor(Actor):
    """
    Epsilon greedy strategy used in DQN
    Epsilon represents the probability of choosing a random action in the environment

    :param critic_encoder: encoder of the critic network
    :param critic_torso: torso of the critic network
    :param critic_head: head of the critic network
    :param start_epsilon: epsilon start value, generally set high to promp random exploration
    :param epsilon_decay: epsilon decay value, epsilon = epsilon * epsilon_decay
    :param min_epsilon: the minimum epsilon value allowed
    :param create_target: whether to create a target network
    :param polyak_coeff: the polyak coefficient for the target network
    :param device: the device to use
    """

    def __init__(
        self,
        critic_encoder: T.nn.Module,
        critic_torso: T.nn.Module,
        critic_head: BaseCriticHead,
        start_epsilon: float = 1,
        epsilon_decay: float = 0.999,
        min_epsilon: float = 0,
        create_target: bool = False,
        polyak_coeff: float = 0.995,
        device: Union[T.device, str] = "auto",
    ):
        super().__init__(
            critic_encoder,
            critic_torso,
            critic_head,
            create_target=create_target,
            polyak_coeff=polyak_coeff,
            device=device,
        )
        self.epsilon = start_epsilon
        self.epsilon_decay = epsilon_decay
        self.min_epsilon = min_epsilon

    def update_epsilon(self) -> None:
        """Update epsilon via epsilon decay"""
        self.epsilon = max(self.epsilon * self.epsilon_decay, self.min_epsilon)

    def forward(self, observations: Tensor) -> T.Tensor:
        q_values = super().forward(observations)
        action_size = q_values.shape[-1]
        trigger = T.rand(1).item()

        if trigger <= self.epsilon:
            actions = T.randint(low=0, high=action_size, size=q_values.shape[:-1])
        else:
            _, actions = T.max(q_values, dim=-1)

        self.update_epsilon()
        return actions


class Dummy(Actor):
    """
    A model which can represent and actor or critic without a nerual network.

    :param space: the individual space
    :param state: optional starting state of the individual
    """

    def __init__(self, space: Space, state: Optional[np.ndarray] = None) -> None:
        # Random network to make compatible with ActorCritic
        super().__init__(
            encoder=IdentityEncoder(),
            torso=MLPEncoder(1, 1),
            head=DummyHead(),
        )
        self.space = space
        self.space_shape = get_space_shape(self.space)
        self.space_range = get_space_range(self.space)
        self.state = np.array(state) if state is not None else np.array(space.sample())

    def set_state(self, state: np.ndarray) -> "Dummy":
        """Set the state of the individual"""
        self.state = np.array(state)
        return self

    def numpy(self) -> np.ndarray:
        """Get the numpy representation of the individual"""
        return self.state

    def forward(self, observation: Tensor) -> np.ndarray:
        return to_torch(self.state)


class ActorCritic(T.nn.Module):
    """
    A basic actor critic network.
    This module is designed flexibly to allow for:
        1. Shared or separate network architectures.
        2. Multiple actors and/or critics defined by `self.actors` and `self.critics`.
        3. A global actor and critic defined as `self.actor` and `self.critic` which is updated as the average of the actor and critic populations.
        4. Handling any target networks embedded in the actor and critic models.

    To define shared layers, simply have the
    actor and critic embedded networks use the same encoder/torso/head
    in memory.

    Separate architecture:
        ```
        encoder_critic = IdentityEncoder()
        encoder_actor = IdentityEncoder()
        torso_critic = MLP([2, 2])
        torso_actor = MLP([2, 2])
        head_actor = DiagGaussianHead(input_shape=2, action_size=1)
        head_critic = ValueHead(input_shape=2, activation_fn=None)
        actor = Actor(encoder=encoder_actor, torso=torso_actor, head=head_actor)
        critic = Critic(encoder=encoder_actor, torso=torso_critic, head=head_critic)
        actor_critic = ActorCritic(actor=actor, critic=critic)
        ```

    Shared architecture:
        ```
        encoder = IdentityEncoder()
        torso = MLP([2, 2])
        head_actor = DiagGaussianHead(input_shape=2, action_size=1)
        head_critic = ValueHead(input_shape=2, activation_fn=None)
        actor = Actor(encoder=encoder, torso=torso, head=head_actor)
        critic = Critic(encoder=encoder, torso=torso, head=head_critic)
        actor_critic = ActorCritic(actor=actor, critic=critic)
        ```

    :param actor: the actor/policy network
    :param critic: the critic network
    :param population_settings: the settings for the actor and critic populations
    """

    def __init__(
        self,
        actor: Actor,
        critic: Critic,
        population_settings: PopulationSettings = PopulationSettings(),
    ) -> None:
        super().__init__()
        self.actor = actor
        self.critic = critic
        self.population_settings = population_settings
        self.num_actors = population_settings.actor_population_size
        self.num_critics = population_settings.critic_population_size
        actor_dist = population_settings.actor_distribution
        critic_dist = population_settings.critic_distribution
        actor_dist = (
            Distribution(actor_dist.lower())
            if isinstance(actor_dist, str)
            else actor_dist
        )
        critic_dist = (
            Distribution(critic_dist.lower())
            if isinstance(critic_dist, str)
            else critic_dist
        )
        self.mean_actor = None
        self.normal_dist_actor = None
        self.actors = self.initialize_population(
            model=actor,
            population_size=self.num_actors,
            population_distribution=actor_dist,
            population_std=population_settings.actor_std,
        )
        self.mean_critic = None
        self.normal_dist_critic = None
        self.critics = self.initialize_population(
            model=critic,
            population_size=self.num_critics,
            population_distribution=critic_dist,
            population_std=population_settings.critic_std,
        )
        if self.actor.model.encoder == self.critic.model.encoder:
            assert self.num_critics == self.num_actors
            for actor, critic in zip(self.actors, self.critics):
                critic.model.encoder = actor.model.encoder
        if self.actor.model.torso == self.critic.model.torso:
            assert self.num_critics == self.num_actors
            for actor, critic in zip(self.actors, self.critics):
                critic.model.torso = actor.model.torso
        if self.actor.model.head == self.critic.model.head:
            assert self.num_critics == self.num_actors
            for actor, critic in zip(self.actors, self.critics):
                critic.model.head = actor.model.head
        self.assign_targets()

    def initialize_population(
        self,
        model: Union[Actor, Critic],
        population_size: int,
        population_distribution: Optional[Distribution],
        population_std: Union[float, np.ndarray] = 1,
    ) -> List[Union[Actor, Critic]]:
        """
        Initialize the population of networks.

        :param model: the model base from which the population is generated.
        :param population_size: the size of the population
        :param population_distribution: the distribution of the population
        :param population_std: the standard deviation of the population if a normal distribution
        """
        if population_distribution is None:
            return [copy.deepcopy(model) for _ in range(population_size)]
        elif population_distribution == Distribution.UNIFORM:
            population = np.random.uniform(
                model.space_range[0],
                model.space_range[1],
                (population_size, *model.space_shape),
            )
        elif population_distribution == Distribution.NORMAL:
            mean = (model.numpy()).astype(np.float32)
            normal_dist = np.random.randn(population_size, *model.space_shape)
            population = mean + (population_std * normal_dist)
            if isinstance(model, Actor):
                self.mean_actor = mean
                self.normal_dist_actor = normal_dist
            elif isinstance(model, Critic):
                self.mean_critic = mean
                self.normal_dist_critic = normal_dist
        else:
            raise ValueError(
                f"The population initialization strategy {population_distribution} is not supported"
            )

        # Discretize and clip population as needed
        if isinstance(model.space, (Discrete, MultiDiscrete)):
            population = np.round(population).astype(np.int32)
        population = np.clip(population, model.space_range[0], model.space_range[1])

        return [copy.deepcopy(model).set_state(ind) for ind in population]

    def numpy_actors(self) -> np.ndarray:
        """Get the numpy representation of the actor population."""
        return np.array([ind.numpy() for ind in self.actors])

    def numpy_critics(self) -> np.ndarray:
        """Get the numpy representation of the critic population."""
        return np.array([ind.numpy() for ind in self.critics])

    def set_actors_state(self, state: np.ndarray) -> "ActorCritic":
        """Set the state of the actors"""
        state = state[np.newaxis] if state.ndim == 1 else state
        [actor.set_state(s) for s, actor in zip(state, self.actors)]
        return self

    def set_critics_state(self, state: np.ndarray) -> "ActorCritic":
        """Set the state of the critics"""
        state = state[np.newaxis] if state.ndim == 1 else state
        [critic.set_state(s) for s, critic in zip(state, self.critics)]
        return self

    def assign_targets(self) -> None:
        """Assign the target parameters"""
        if self.actor.target is not None:
            [actor.assign_targets() for actor in self.actors]
        if self.critic.target is not None:
            [critic.assign_targets() for critic in self.critics]

    def update_targets(self) -> None:
        """Update the target parameters"""
        # target_params = polyak_coeff * target_params + (1 - polyak_coeff) * online_params
        if self.actor.target is not None:
            [actor.update_targets() for actor in self.actors]
        if self.critic.target is not None:
            [critic.update_targets() for critic in self.critics]

    def update_global(self) -> None:
        """Update global networks"""
        if self.num_actors == 1:
            self.actor = self.actors[0]
        else:
            actor_states = [actor.state for actor in self.actors]
            self.actor.set_state(np.mean(actor_states, axis=0))

        if self.num_critics == 1:
            self.critic = self.critics[0]
        else:
            critic_states = [critic.state for critic in self.critics]
            self.critic.set_state(np.mean(critic_states, axis=0))

    def action_distribution(
        self, observations: Tensor
    ) -> Optional[T.distributions.Distribution]:
        """Get the population action distributions, returns None if deterministic"""
        if self.num_actors == 1:
            return self.actors[0].action_distribution(observations)
        distributions = [
            actor.action_distribution(obs)
            for actor, obs in zip(self.actors, observations)
        ]
        if all(dist is None for dist in distributions):
            return None
        elif isinstance(self.actor.model.head, CategoricalHead):
            logits = T.stack([dist.logits for dist in distributions])
            return T.distributions.Categorical(logits=logits)
        elif isinstance(self.actor.model.head, DiagGaussianHead):
            means = T.stack([dist.mean for dist in distributions])
            stds = T.stack([dist.stddev for dist in distributions])
            return T.distributions.Normal(means, stds)

    def forward_target_critics(
        self, observations: Tensor, actions: Optional[Tensor] = None
    ) -> T.Tensor:
        """Get the population target critic outputs"""
        if self.num_critics == 1:
            return self.critics[0].forward_target(observations, actions)
        elif actions is None:
            return T.stack(
                [
                    critic.forward_target(obs).to(critic.device)
                    for critic, obs in zip(self.critics, observations)
                ]
            )
        return T.stack(
            [
                critic.forward_target(obs, action).to(critic.device)
                for critic, obs, action in zip(self.critics, observations, actions)
            ]
        )

    def forward_target_actors(self, observations: Tensor) -> T.Tensor:
        """Get the population target actor outputs"""
        if self.num_actors == 1:
            return self.actors[0].forward_target(observations)
        return T.stack(
            [
                actor.forward_target(obs).to(actor.device)
                for actor, obs in zip(self.actors, observations)
            ]
        )

    def forward_critics(
        self, observations: Tensor, actions: Optional[Tensor] = None
    ) -> T.Tensor:
        """Get the population online critic outputs"""
        if self.num_critics == 1:
            return self.critics[0](observations, actions)
        elif actions is None:
            return T.stack(
                [
                    critic(obs).to(critic.device)
                    for critic, obs in zip(self.critics, observations)
                ]
            )
        return T.stack(
            [
                critic(obs, action).to(critic.device)
                for critic, obs, action in zip(self.critics, observations, actions)
            ]
        )

    def forward(self, observations: Tensor) -> T.Tensor:
        """The default forward pass retrieves the population online actor outputs"""
        if self.num_actors == 1:
            return self.actors[0](observations)
        return T.stack(
            [
                actor(obs).to(actor.device)
                for actor, obs in zip(self.actors, observations)
            ]
        )

    def predict_distribution(
        self, observations: Tensor
    ) -> Optional[T.distributions.Distribution]:
        """Get the global network action distribution, returns None if deterministic"""
        return self.actor.action_distribution(observations)

    def predict(self, observations: Tensor) -> T.Tensor:
        """
        Get the global network action prediction

        :param observations: The state observations
        :return: The action prediction
        """
        return self.actor(observations)

    def predict_critic(
        self, observations: Tensor, actions: Optional[Tensor] = None
    ) -> T.Tensor:
        """
        Predict the critic value for a given state and optional action

        :param observations: The state observations
        :param actions: The actions
        :return: The critic value
        """
        return self.critic(observations, actions)



from abc import ABC, abstractmethod

from pearll.common.logging_ import Logger
from pearll.models.actor_critics import ActorCritic


class BaseCallback(ABC):
    """
    Base class for callback.
    :param logger:
    """

    def __init__(self, logger: Logger, model: ActorCritic) -> None:
        self.n_calls = 0
        self.step = 0
        self.logger = logger
        self.model = model

    @abstractmethod
    def _on_step(self) -> bool:
        """
        :return: If the callback returns False, training is aborted early.
        """

    def on_step(self, step: int) -> bool:
        """
        This method will be called by the model after each call to ``step_env()``.
        For child callback (of an ``EventCallback``), this will be called
        when the event is triggered.
        :return: If the callback returns False, training is aborted early.
        """
        self.n_calls += 1
        self.step = step

        return self._on_step()



from abc import ABC, abstractmethod
from typing import Iterator, Type

import torch as T
from torch.nn import functional as F

from pearll.common.type_aliases import UpdaterLog
from pearll.models.actor_critics import Model


class BaseDeepUpdater(ABC):
    """
    Base class for updating a deep environment model.
    """

    def __init__(
        self,
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        self.optimizer_class = optimizer_class
        self.max_grad = max_grad

    def run_optimizer(
        self,
        optimizer: T.optim.Optimizer,
        loss: T.Tensor,
        model_parameters: Iterator[T.nn.Parameter],
    ) -> None:
        """Run an optimization step"""
        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        if self.max_grad > 0:
            T.nn.utils.clip_grad_norm_(model_parameters, self.max_grad)
        optimizer.step()

    @abstractmethod
    def __call__(
        self,
        model: Model,
        observations: T.Tensor,
        actions: T.Tensor,
        targets: T.Tensor,
        learning_rate: float = 0.001,
    ) -> UpdaterLog:
        """
        Run an optimization step

        :param model: The model to update (observation, reward or done models)
        :param observations: The input observations
        :param actions: The input actions
        :param targets: The targets to regress against
        :param learning_rate: The learning rate to use
        """


class DeepRegression(BaseDeepUpdater):
    """
    Update a deep model using a deep regression algorithm.

    :param loss_class: The loss class to use e.g. MSE
    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        loss_class: T.nn.Module = T.nn.MSELoss(),
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        super().__init__(optimizer_class, max_grad)
        self.loss_class = loss_class

    def __call__(
        self,
        model: Model,
        observations: T.Tensor,
        actions: T.Tensor,
        targets: T.Tensor,
        learning_rate: float = 0.001,
        mode: str = "auto",
    ) -> UpdaterLog:
        """
        Run an optimization step

        :param model: The model to update (observation, reward or done models)
        :param observations: The input observations
        :param actions: The input actions
        :param targets: The targets to regress against
        :param learning_rate: The learning rate to use
        :param mode: The mode to use, defaults to auto. If set to anything else, no processing
            will be done on the targets and predictions for different loss functions.
        """
        params = model.parameters()
        predictions = model(observations, actions)

        if mode == "auto":
            # Data processing for categorical losses; assumes categorical target.
            if isinstance(self.loss_class, (T.nn.CrossEntropyLoss, T.nn.BCELoss)):
                torso_output = model.torso(model.encoder(observations, actions))
                predictions = super(type(model.head), model.head).forward(torso_output)
                targets = (
                    F.one_hot(targets.long(), predictions.shape[-1]).squeeze().float()
                )

        optimizer = self.optimizer_class(params, lr=learning_rate)
        loss = self.loss_class(predictions, targets)
        self.run_optimizer(optimizer, loss, params)

        return UpdaterLog(loss=loss.detach().item())



import copy
from dataclasses import dataclass
from functools import partial
from typing import Callable, List, Optional, Type

import numpy as np
import torch as T
from gym import Env
from gym.vector.vector_env import VectorEnv

from pearll.agents.base_agents import BaseAgent
from pearll.buffers import ReplayBuffer
from pearll.buffers.base_buffer import BaseBuffer
from pearll.callbacks.base_callback import BaseCallback
from pearll.common.type_aliases import Log
from pearll.common.utils import filter_rewards, get_space_shape, to_numpy
from pearll.explorers import BaseExplorer
from pearll.models import Actor, ActorCritic, Critic
from pearll.models.encoders import IdentityEncoder, MLPEncoder
from pearll.models.heads import ContinuousQHead, DeterministicHead
from pearll.models.torsos import MLP
from pearll.models.utils import get_mlp_size
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    MiscellaneousSettings,
    OptimizerSettings,
    PopulationSettings,
    Settings,
)
from pearll.signal_processing import (
    crossover_operators,
    return_estimators,
    selection_operators,
)
from pearll.updaters.critics import BaseCriticUpdater, ContinuousQRegression
from pearll.updaters.evolution import BaseEvolutionUpdater, GeneticUpdater


def get_default_model(env: Env) -> ActorCritic:
    action_shape = get_space_shape(env.single_action_space)
    observation_shape = get_space_shape(env.single_observation_space)
    action_size = get_mlp_size(action_shape)
    observation_size = get_mlp_size(observation_shape)
    encoder_actor = IdentityEncoder()
    encoder_critic = MLPEncoder(observation_size + action_size, observation_size)
    torso = MLP(layer_sizes=[observation_size, 40, 30], activation_fn=T.nn.ReLU)
    head_actor = DeterministicHead(
        input_shape=30, action_shape=action_shape, activation_fn=T.nn.Tanh
    )
    head_critic = ContinuousQHead(input_shape=30)
    return ActorCritic(
        actor=Actor(
            encoder=encoder_actor,
            torso=torso,
            head=head_actor,
            create_target=True,
        ),
        critic=Critic(
            encoder=encoder_critic,
            torso=torso,
            head=head_critic,
            create_target=True,
        ),
        population_settings=PopulationSettings(
            actor_population_size=env.num_envs,
            critic_population_size=env.num_envs,
            actor_distribution="normal",
            actor_std=2,
        ),
    )


@dataclass
class NaiveSelectionSettings(Settings):
    ratio: float = 0.5


class CEM_RL(BaseAgent):
    """
    CEM-RL Algorithm

    :param env: the gym-like environment to be used
    :param eval_env: the environment to be used for evaluating agent before evolutionary update
    :param model: the neural network model
    :param td_gamma: trajectory discount factor
    :param value_coeff: value loss weight
    :param actor_updater_class: actor updater class
    :param actor_optimizer_settings: actor optimizer settings
    :param critic_updater_class: critic updater class
    :param critic_optimizer_settings: critic optimizer settings
    :param buffer_class: the buffer class for storing and sampling trajectories
    :param buffer_settings: settings for the buffer
    :param action_explorer_class: the explorer class for random search at beginning of training and
        adding noise to actions
    :param explorer_settings: settings for the action explorer
    :param callbacks: an optional list of callbacks (e.g. if you want to save the model)
    :param callback_settings: settings for callbacks
    :param logger_settings: settings for the logger
    :param misc_settings: settings for miscellaneous parameters
    """

    def __init__(
        self,
        env: VectorEnv,
        eval_env: VectorEnv,
        model: Optional[ActorCritic],
        td_gamma: float = 0.99,
        value_coeff: float = 0.5,
        actor_updater_class: Type[BaseEvolutionUpdater] = GeneticUpdater,
        selection_operator: Callable = selection_operators.naive_selection,
        selection_settings: Settings = NaiveSelectionSettings(),
        crossover_operator: Callable = crossover_operators.fit_gaussian,
        critic_updater_class: Type[BaseCriticUpdater] = ContinuousQRegression,
        critic_optimizer_settings: OptimizerSettings = OptimizerSettings(),
        buffer_class: Type[BaseBuffer] = ReplayBuffer,
        buffer_settings: BufferSettings = BufferSettings(),
        action_explorer_class: Type[BaseExplorer] = BaseExplorer,
        explorer_settings: ExplorerSettings = ExplorerSettings(start_steps=0),
        callbacks: Optional[List[Type[BaseCallback]]] = None,
        callback_settings: Optional[List[Settings]] = None,
        logger_settings: LoggerSettings = LoggerSettings(),
        misc_settings: MiscellaneousSettings = MiscellaneousSettings(),
    ) -> None:
        model = model or get_default_model(env)
        super().__init__(
            env,
            model,
            action_explorer_class=action_explorer_class,
            explorer_settings=explorer_settings,
            buffer_class=buffer_class,
            buffer_settings=buffer_settings,
            logger_settings=logger_settings,
            callbacks=callbacks,
            callback_settings=callback_settings,
            misc_settings=misc_settings,
        )
        self.eval_env = eval_env
        self.actor_updater = actor_updater_class(self.model)
        self.critic_updater = critic_updater_class(
            loss_class=critic_optimizer_settings.loss_class,
            optimizer_class=critic_optimizer_settings.optimizer_class,
            max_grad=critic_optimizer_settings.max_grad,
        )

        self.critic_optimizer_settings = critic_optimizer_settings
        self.value_coeff = value_coeff
        self.td_gamma = td_gamma
        self.selection_operator = partial(
            selection_operator, **selection_settings.filter_none()
        )
        self.crossover_operator = partial(
            crossover_operator, population_shape=self.model.numpy_actors().shape
        )

    def _fit(self, batch_size: int, actor_epochs: int = 1, critic_epochs: int = 1):
        critic_losses = np.zeros(critic_epochs)
        divergences = np.zeros(actor_epochs)
        entropies = np.zeros(actor_epochs)

        model_copy = copy.deepcopy(self.model)

        # Train critic for critic_epochs
        for i in range(critic_epochs):
            trajectories = self.buffer.sample(batch_size=batch_size)
            with T.no_grad():
                next_actions = self.model.forward_target_actors(
                    trajectories.next_observations
                )
                next_q_values = self.model.forward_target_critics(
                    trajectories.next_observations, next_actions
                )
                target_q_values = return_estimators.TD_zero(
                    trajectories.rewards,
                    next_q_values,
                    trajectories.dones,
                    gamma=self.td_gamma,
                )
            critic_log = self.critic_updater(
                self.model,
                trajectories.observations,
                trajectories.actions,
                target_q_values,
                learning_rate=self.critic_optimizer_settings.learning_rate,
                loss_coeff=self.value_coeff,
            )
            critic_losses[i] = critic_log.loss

        # Reset half the actors to their state before being updated by the critic updater
        half_actors = self.model.num_actors // 2
        actors_no_update = model_copy.numpy_actors()[:half_actors]
        actors_with_update = self.model.numpy_actors()[half_actors:]
        new_state = np.concatenate([actors_no_update, actors_with_update], axis=0)
        self.model.set_actors_state(new_state)

        # Evaluate new model
        episode_dones = [False for _ in range(self.eval_env.num_envs)]
        observation = self.eval_env.reset()
        episode_length = 0
        while not np.all(episode_dones):
            action = to_numpy(self.model(observation))
            next_observation, reward, done, _ = self.eval_env.step(action)
            self.buffer.add_trajectory(
                observation, action, reward, next_observation, done
            )
            episode_length += 1
            observation = next_observation
            episode_dones = np.logical_or(episode_dones, done)

        trajectories = self.buffer.last(episode_length, flatten_env=False)
        rewards = trajectories.rewards.squeeze()
        rewards = filter_rewards(rewards, trajectories.dones.squeeze())
        if rewards.ndim > 1:
            rewards = rewards.sum(axis=-1)

        # Train actor for actor_epochs
        for i in range(actor_epochs):
            actor_log = self.actor_updater(
                rewards=rewards,
                selection_operator=self.selection_operator,
                crossover_operator=self.crossover_operator,
                elitism=0,
            )
            divergences[i] = actor_log.divergence
            entropies[i] = actor_log.entropy

        # Update target networks
        self.model.update_targets()

        return Log(
            critic_loss=np.mean(critic_losses),
            divergence=np.mean(divergences),
            entropy=np.mean(entropies),
        )



from typing import List, Optional, Type

import torch as T

from pearll.common.type_aliases import Tensor


class MLP(T.nn.Module):
    def __init__(
        self, layer_sizes: List[int], activation_fn: Optional[Type[T.nn.Module]] = None
    ):
        super().__init__()

        layers = []
        for i, size in enumerate(layer_sizes[:-1]):
            layers += [T.nn.Linear(size, layer_sizes[i + 1])]
            if activation_fn:
                layers += [activation_fn()]
        self.model = T.nn.Sequential(*layers)

    def forward(self, inputs: Tensor) -> T.Tensor:
        return self.model(inputs)



from pearll.updaters.actors import BaseActorUpdater
from pearll.updaters.critics import BaseCriticUpdater
from pearll.updaters.environment import BaseDeepUpdater
from pearll.updaters.evolution import BaseEvolutionUpdater

__all__ = [
    "BaseActorUpdater",
    "BaseCriticUpdater",
    "BaseEvolutionUpdater",
    "BaseDeepUpdater",
]



"""This module holds settings objects to configure the other modules"""

import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional, Tuple, Type, Union

import numpy as np
import torch as T
from torch.optim.optimizer import Optimizer

from pearll.common.enumerations import Distribution


@dataclass
class Settings:
    """Base class for settings objects"""

    def filter_none(self) -> Dict[str, Any]:
        return {k: v for k, v in self.__dict__.items() if v is not None}


@dataclass
class MiscellaneousSettings(Settings):
    """
    Miscellaneous settings for base agent

    :param seed: random seed
    :param device: device to use for computations
    :param render: whether to render the environment
    """

    device: str = "auto"
    render: bool = False
    seed: Optional[int] = None


@dataclass
class OptimizerSettings(Settings):
    """
    Settings for the model optimizers

    :param loss_class: optional surrogate loss class to use for the optimizer
    :param optimizer_class: class of optimizer algorithm to use
    :param learning_rate: optimizer learning rate
    :param max_grad: maximum gradient for gradient clipping
    """

    loss_class: Optional[T.nn.Module] = T.nn.MSELoss()
    optimizer_class: Type[Optimizer] = T.optim.Adam
    learning_rate: Optional[float] = 1e-3
    max_grad: float = 0.5


@dataclass
class PopulationSettings(Settings):
    """
    Settings for the population initializer

    :param actor_population_size: number of actors in the population
    :param critic_population_size: number of critics in the population
    :param actor_distribution: distribution of the actor population
    :param critic_distribution: distribution of the critic population
    :param actor_std: standard deviation of the actor population if normally distributed
    :param critic_std: standard deviation of the critic population if normally distributed
    """

    actor_population_size: int = 1
    critic_population_size: int = 1
    actor_distribution: Optional[Union[str, Distribution]] = None
    critic_distribution: Optional[Union[str, Distribution]] = None
    actor_std: Optional[Union[float, np.ndarray]] = 1
    critic_std: Optional[Union[float, np.ndarray]] = 1


@dataclass
class ExplorerSettings(Settings):
    """
    Settings for the action explorer

    :param start_steps: number of steps at the start to randomly sample actions (encourages exploration)
    :param scale: std of noise to add to actions (not always applicable)
    """

    start_steps: int = 1000
    scale: Optional[float] = None


@dataclass
class BufferSettings(Settings):
    """
    Settings for buffers

    :buffer_size: max number of transitions to store at once in each environment
    """

    buffer_size: int = int(1e6)


@dataclass
class LoggerSettings(Settings):
    """
    Settings for the Logger

    :param tensorboard_log_path: path to store the tensorboard log
    :param file_handler_level: logging level for the file log
    :param stream_handler_level: logging level for the streaming log
    :param verbose: whether to record any logs at all
    """

    tensorboard_log_path: Optional[str] = None
    log_frequency: Tuple[str, int] = ("episode", 1)
    file_handler_level: int = logging.DEBUG
    stream_handler_level: int = logging.INFO
    verbose: bool = True


@dataclass
class MutationSettings(Settings):
    """
    Settings for the mutation process. Extend this class to add params for each mutation method.

    :param mutation_rate: probability of mutation for each individual
    :param mutation_std: optional standard deviation of the mutation distribution
    """

    mutation_rate: float = 0.1
    mutation_std: Optional[float] = None



from typing import Union

import numpy as np
import torch as T
from gym import Env

from pearll.buffers.base_buffer import BaseBuffer
from pearll.common.enumerations import TrajectoryType
from pearll.common.type_aliases import Trajectories


class RolloutBuffer(BaseBuffer):
    """
    Rollout buffer handles sample collection and processing for on-policy algorithms.

    :param env: the environment
    :param buffer_size: max number of elements in the buffer
    """

    def __init__(
        self,
        env: Env,
        buffer_size: int,
        device: Union[str, T.device] = "auto",
    ) -> None:
        super().__init__(
            env,
            buffer_size,
            device,
        )
        self.next_observations = np.zeros(
            (self.buffer_size,) + self.obs_shape,
            dtype=env.observation_space.dtype,
        )
        self._check_system_memory(
            self.observations,
            self.actions,
            self.rewards,
            self.dones,
            self.next_observations,
        )

    def reset(self) -> None:
        super().reset()
        self.next_observations = np.zeros(
            (self.buffer_size,) + self.obs_shape,
            dtype=self.env.observation_space.dtype,
        )

    def add_trajectory(
        self,
        observation: np.ndarray,
        action: Union[np.ndarray, int],
        reward: Union[float, np.ndarray],
        next_observation: np.ndarray,
        done: Union[bool, np.ndarray],
    ) -> None:
        self.observations[self.pos] = observation
        self.next_observations[self.pos] = next_observation
        self.actions[self.pos] = np.array(action).reshape(*self.actions.shape[1:])
        self.rewards[self.pos] = np.array(reward).reshape(*self.rewards.shape[1:])
        self.dones[self.pos] = np.array(done).reshape(*self.dones.shape[1:])

        self.pos += 1
        if self.pos == self.buffer_size:
            self.full = True
            self.pos = 0

    def sample(
        self,
        batch_size: int,
        flatten_env: bool = False,
        dtype: Union[str, TrajectoryType] = "torch",
    ) -> Trajectories:
        if self.full:
            assert (
                batch_size <= self.buffer_size
            ), f"Requesting {batch_size} samples when only {self.buffer_size} samples have been collected"
            upper_bound = self.buffer_size
        else:
            assert (
                batch_size <= self.pos
            ), f"Requesting {batch_size} samples when only {self.pos} samples have been collected"
            upper_bound = self.pos
        start_idx = np.random.randint(0, (upper_bound + 1) - batch_size)
        last_idx = start_idx + batch_size

        observations = self.observations[start_idx:last_idx]
        actions = self.actions[start_idx:last_idx]
        rewards = self.rewards[start_idx:last_idx]
        next_observations = self.next_observations[start_idx:last_idx]
        dones = self.dones[start_idx:last_idx]

        return self._transform_samples(
            flatten_env, dtype, observations, actions, rewards, next_observations, dones
        )

    def last(
        self,
        batch_size: int,
        flatten_env: bool = False,
        dtype: Union[str, TrajectoryType] = "torch",
    ) -> Trajectories:
        assert batch_size <= self.buffer_size

        start_idx = self.pos - batch_size
        if start_idx < 0:
            batch_inds = np.concatenate((np.arange(start_idx, 0), np.arange(self.pos)))
        else:
            batch_inds = np.arange(start_idx, self.pos)

        observations = self.observations[batch_inds]
        actions = self.actions[batch_inds]
        rewards = self.rewards[batch_inds]
        next_observations = self.next_observations[batch_inds]
        dones = self.dones[batch_inds]

        return self._transform_samples(
            flatten_env, dtype, observations, actions, rewards, next_observations, dones
        )

    def all(
        self, flatten_env: bool = False, dtype: Union[str, TrajectoryType] = "torch"
    ) -> Trajectories:
        return self._transform_samples(
            flatten_env=flatten_env,
            dtype=dtype,
            observations=self.observations[: self.pos],
            actions=self.actions[: self.pos],
            rewards=self.rewards[: self.pos],
            next_observations=self.next_observations[: self.pos],
            dones=self.dones[: self.pos],
        )



"""Methods for selecting individuals in a population to evolve for the next algorithm iteration"""

import numpy as np


def naive_selection(
    population: np.ndarray, fitness_scores: np.ndarray, ratio: float = 0.5
) -> np.ndarray:
    """
    Selects individuals for the next algorithm iteration using naive selection.

    :param population: the population of individuals to select from
    :param fitness_scores: the fitness scores of the individuals in the population
    :param ratio: the ratio of the number of individuals to select from the population
    :return: the selected individuals
    """
    num_selected = max(1, int(ratio * population.shape[0]))
    return population[np.argpartition(fitness_scores, -num_selected)[-num_selected:]]


def tournament_selection(
    population: np.ndarray,
    fitness_scores: np.ndarray,
    tournament_size: int = 2,
    probability: float = 0.8,
) -> np.ndarray:
    """
    Selects individuals for the next algorithm iteration using tournament selection.
    https://www.geeksforgeeks.org/tournament-selection-ga/

    :param population: the population of individuals to select from
    :param fitness_scores: the fitness scores of the individuals in the population
    :param tournament_size: the number of individuals to select from the population
    :param probability: the probability of selecting the best individual for each tournament
    :return: the selected individuals
    """
    # Need a combined data structure to keep track of the fitness scores of the individuals after sorting
    combined_data = {
        fitness_score: individual
        for individual, fitness_score in zip(population, fitness_scores)
    }

    # Sort the combined data structure to get sorted fitness scores and population
    sorted_fitness = np.flip(sorted(combined_data))

    # Calculate probabilities of selecting an individual for tournament
    probabilities = np.array(
        [probability * ((1 - probability) ** i) for i in range(fitness_scores.shape[0])]
    )
    # Ensure valid probability mass function (sums to 1)
    residual = 1 - np.sum(probabilities)
    probabilities[0] += residual

    # Create the tournament
    tournament_fitness = np.random.choice(
        sorted_fitness,
        size=(population.shape[0], tournament_size),
        p=probabilities,
    )
    winning_fitness = np.max(tournament_fitness, axis=1)
    selected_individuals = np.array(
        [combined_data[fitness_score] for fitness_score in winning_fitness]
    )

    return selected_individuals


def roulette_selection(
    population: np.ndarray, fitness_scores: np.ndarray
) -> np.ndarray:
    """
    Selects individuals for the next algorithm iteration using roulette selection.

    :param population: the population of individuals to select from
    :param fitness_scores: the fitness scores of the individuals in the population
    :return: the selected individuals
    """
    # Calculate the probabilities of selecting an individual
    fitness_scores = fitness_scores / np.sum(fitness_scores)

    # Select individuals based on the probabilities
    selected_indices = np.random.choice(
        population.shape[0], size=population.shape[0], p=fitness_scores
    )
    return population[selected_indices]



from typing import List, Optional, Type, Union

import numpy as np
import torch as T
from gym import Env

from pearll.agents.base_agents import BaseAgent
from pearll.buffers import ReplayBuffer
from pearll.buffers.base_buffer import BaseBuffer
from pearll.callbacks.base_callback import BaseCallback
from pearll.common.type_aliases import Log
from pearll.common.utils import get_space_shape, to_numpy
from pearll.explorers import BaseExplorer, GaussianExplorer
from pearll.models import Actor, ActorCritic, Critic
from pearll.models.encoders import IdentityEncoder
from pearll.models.heads import ContinuousQHead, DeterministicHead
from pearll.models.torsos import MLP
from pearll.models.utils import get_mlp_size
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    MiscellaneousSettings,
    OptimizerSettings,
    Settings,
)
from pearll.signal_processing.return_estimators import TD_zero
from pearll.updaters.actors import BaseActorUpdater, DeterministicPolicyGradient
from pearll.updaters.critics import BaseCriticUpdater, ContinuousQRegression


def get_default_model(env: Env) -> ActorCritic:
    action_shape = get_space_shape(env.action_space)
    observation_shape = get_space_shape(env.observation_space)
    action_size = get_mlp_size(action_shape)
    observation_size = get_mlp_size(observation_shape)
    encoder_actor = IdentityEncoder()
    encoder_critic = IdentityEncoder()
    torso_actor = MLP(layer_sizes=[observation_size, 400, 300], activation_fn=T.nn.ReLU)
    torso_critic = MLP(
        layer_sizes=[observation_size + action_size, 400, 300], activation_fn=T.nn.ReLU
    )
    head_actor = DeterministicHead(
        input_shape=300, action_shape=action_shape, activation_fn=T.nn.Tanh
    )
    head_critic = ContinuousQHead(input_shape=300)
    return ActorCritic(
        actor=Actor(
            encoder=encoder_actor,
            torso=torso_actor,
            head=head_actor,
            create_target=True,
        ),
        critic=Critic(
            encoder=encoder_critic,
            torso=torso_critic,
            head=head_critic,
            create_target=True,
        ),
    )


class DDPG(BaseAgent):
    """
    DDPG Algorithm

    :param env: the gym-like environment to be used
    :param model: the neural network model
    :param td_gamma: trajectory discount factor
    :param value_coeff: value loss weight
    :param actor_updater_class: actor updater class
    :param actor_optimizer_settings: actor optimizer settings
    :param critic_updater_class: critic updater class
    :param critic_optimizer_settings: critic optimizer settings
    :param buffer_class: the buffer class for storing and sampling trajectories
    :param buffer_settings: settings for the buffer
    :param action_explorer_class: the explorer class for random search at beginning of training and
        adding noise to actions
    :param explorer_settings: settings for the action explorer
    :param callbacks: an optional list of callbacks (e.g. if you want to save the model)
    :param callback_settings: settings for callbacks
    :param logger_settings: settings for the logger
    :param misc_settings: settings for miscellaneous parameters
    """

    def __init__(
        self,
        env: Env,
        model: Optional[ActorCritic],
        td_gamma: float = 0.99,
        value_coeff: float = 0.5,
        actor_updater_class: Type[BaseActorUpdater] = DeterministicPolicyGradient,
        actor_optimizer_settings: OptimizerSettings = OptimizerSettings(),
        critic_updater_class: Type[BaseCriticUpdater] = ContinuousQRegression,
        critic_optimizer_settings: OptimizerSettings = OptimizerSettings(),
        buffer_class: Type[BaseBuffer] = ReplayBuffer,
        buffer_settings: BufferSettings = BufferSettings(),
        action_explorer_class: Type[BaseExplorer] = GaussianExplorer,
        explorer_settings: ExplorerSettings = ExplorerSettings(
            start_steps=1000, scale=0.1
        ),
        callbacks: Optional[List[Type[BaseCallback]]] = None,
        callback_settings: Optional[List[Settings]] = None,
        logger_settings: LoggerSettings = LoggerSettings(),
        misc_settings: MiscellaneousSettings = MiscellaneousSettings(),
    ) -> None:
        model = model or get_default_model(env)
        super().__init__(
            env,
            model,
            action_explorer_class=action_explorer_class,
            explorer_settings=explorer_settings,
            buffer_class=buffer_class,
            buffer_settings=buffer_settings,
            logger_settings=logger_settings,
            callbacks=callbacks,
            callback_settings=callback_settings,
            misc_settings=misc_settings,
        )
        self.actor_updater = actor_updater_class(
            optimizer_class=actor_optimizer_settings.optimizer_class,
            max_grad=actor_optimizer_settings.max_grad,
        )
        self.critic_updater = critic_updater_class(
            loss_class=critic_optimizer_settings.loss_class,
            optimizer_class=critic_optimizer_settings.optimizer_class,
            max_grad=critic_optimizer_settings.max_grad,
        )

        self.actor_optimizer_settings = actor_optimizer_settings
        self.critic_optimizer_settings = critic_optimizer_settings
        self.value_coeff = value_coeff
        self.td_gamma = td_gamma

    def _fit(self, batch_size: int, actor_epochs: int = 1, critic_epochs: int = 1):
        critic_losses = np.zeros(shape=(critic_epochs))
        actor_losses = np.zeros(shape=(actor_epochs))
        # Train critic for critic_epochs
        for i in range(critic_epochs):
            trajectories = self.buffer.sample(batch_size=batch_size)
            with T.no_grad():
                next_actions = self.model.forward_target_actors(
                    trajectories.next_observations
                )
                next_q_values = self.model.forward_target_critics(
                    trajectories.next_observations, next_actions
                )
                target_q_values = TD_zero(
                    trajectories.rewards,
                    next_q_values,
                    trajectories.dones,
                    gamma=self.td_gamma,
                )
            critic_log = self.critic_updater(
                self.model,
                trajectories.observations,
                trajectories.actions,
                target_q_values,
                learning_rate=self.critic_optimizer_settings.learning_rate,
                loss_coeff=self.value_coeff,
            )
            critic_losses[i] = critic_log.loss

        # Train actor for actor_epochs
        for i in range(actor_epochs):
            trajectories = self.buffer.sample(batch_size=batch_size)
            actor_log = self.actor_updater(self.model, trajectories.observations)
            actor_losses[i] = actor_log.loss

        # Update target networks
        self.model.update_targets()

        return Log(
            actor_loss=np.mean(actor_losses),
            critic_loss=np.mean(critic_losses),
        )



from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional, Union

import numpy as np
import torch as T
from gym import Space

Tensor = Union[np.ndarray, T.Tensor]
Observation = Union[np.ndarray, Dict[str, np.ndarray]]

SelectionFunc = Callable[[np.ndarray, np.ndarray], np.ndarray]
CrossoverFunc = Callable[[np.ndarray], np.ndarray]
MutationFunc = Callable[[np.ndarray, Space], np.ndarray]
ObservationFunc = Callable[[Any, Any], Any]
RewardFunc = Callable[[Any, Any], float]
DoneFunc = Callable[[Any, Any], bool]


@dataclass
class Trajectories:
    """Sample trajectory data needed for algorithms"""

    observations: Tensor
    actions: Tensor
    rewards: Tensor
    next_observations: Tensor
    dones: Tensor


@dataclass
class DictTrajectories:
    """Sample trajectory data with dictionary observations needed for algorithms"""

    observations: Dict[str, Tensor]
    actions: Tensor
    rewards: Tensor
    next_observations: Dict[str, Tensor]
    dones: Tensor


@dataclass
class UpdaterLog:
    """Log to see updater metrics for algorithms"""

    loss: Optional[float] = None
    divergence: Optional[float] = None
    entropy: Optional[float] = None


@dataclass
class Log:
    """
    Log to see training progress

    :param reward: reward received
    :param actor_loss: actor network loss
    :param critic_loss: critic network loss
    :divergence: divergence of policy
    :entropy: entropy of policy
    """

    reward: float = 0
    actor_loss: Optional[float] = None
    critic_loss: Optional[float] = None
    divergence: Optional[float] = None
    entropy: Optional[float] = None



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Type, Union

import numpy as np
import torch as T
from gym import Env
from gym.vector import VectorEnv

from pearll.buffers.base_buffer import BaseBuffer
from pearll.callbacks.base_callback import BaseCallback
from pearll.common.enumerations import FrequencyType
from pearll.common.logging_ import Logger
from pearll.common.type_aliases import Log, Observation, Tensor, Trajectories
from pearll.common.utils import get_device, set_seed
from pearll.explorers.base_explorer import BaseExplorer
from pearll.models.actor_critics import ActorCritic
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    MiscellaneousSettings,
    Settings,
)


class BaseAgent(ABC):
    """
    The BaseAgent class is given to handle all the stuff around either RL or EC algorithms.
    It's recommended to inherit this class when implementing your own agents. You'll need
    to implement the _fit() abstract method and override the __init__ to add updaters along
    with their settings.

    See the example agents already done for guidance and settings.py for settings objects
    that can be used.

    :param env: the gym-like environment to be used
    :param model: the neural network model
    :param buffer_class: the buffer class for storing and sampling trajectories
    :param buffer_settings: settings for the buffer
    :param action_explorer_class: the explorer class for random search at beginning of training and
        adding noise to actions
    :param explorer settings: settings for the action explorer
    :param callbacks: an optional list of callbacks (e.g. if you want to save the model)
    :param callback_settings: settings for callbacks
    :param logger_settings: settings for the logger
    :param misc_settings: settings for miscellaneous parameters
    """

    def __init__(
        self,
        env: Env,
        model: ActorCritic,
        buffer_class: Type[BaseBuffer] = BaseBuffer,
        buffer_settings: BufferSettings = BufferSettings(),
        action_explorer_class: Type[BaseExplorer] = BaseExplorer,
        explorer_settings: ExplorerSettings = ExplorerSettings(),
        callbacks: Optional[List[Type[BaseCallback]]] = None,
        callback_settings: Optional[List[Settings]] = None,
        logger_settings: LoggerSettings = LoggerSettings(),
        misc_settings: MiscellaneousSettings = MiscellaneousSettings(),
    ) -> None:
        self.env = env
        self.model = model
        self.render = misc_settings.render
        explorer_settings = explorer_settings.filter_none()
        self.action_explorer = action_explorer_class(
            action_space=env.action_space, **explorer_settings
        )
        buffer_settings = buffer_settings.filter_none()
        self.buffer = buffer_class(
            env=env, device=misc_settings.device, **buffer_settings
        )
        self.step = 0
        self.episode = 0
        self.done = False  # Flag terminate training
        self.logger = Logger(
            tensorboard_log_path=logger_settings.tensorboard_log_path,
            file_handler_level=logger_settings.file_handler_level,
            stream_handler_level=logger_settings.stream_handler_level,
            verbose=logger_settings.verbose,
            num_envs=env.num_envs if isinstance(env, VectorEnv) else 1,
        )
        self.log_frequency = (
            FrequencyType(logger_settings.log_frequency[0].lower()),
            logger_settings.log_frequency[1],
        )
        if callbacks is not None:
            assert len(callbacks) == len(
                callback_settings
            ), "There should be a CallbackSetting object for each callback"
            callback_settings = [setting.filter_none() for setting in callback_settings]
            self.callbacks = [
                callback(logger=self.logger, model=self.model, **settings)
                for callback, settings in zip(callbacks, callback_settings)
            ]
        else:
            self.callbacks = None

        device = get_device(misc_settings.device)
        self.logger.info(f"Using device {device}")

        if misc_settings.seed is not None:
            self.logger.info(f"Using seed {misc_settings.seed}")
            set_seed(misc_settings.seed, self.env)

    @T.no_grad()
    def predict(self, observations: Union[Tensor, Dict[str, Tensor]]) -> T.Tensor:
        """Run the agent actor model"""
        self.model.eval()
        return self.model.predict(observations)

    @T.no_grad()
    def action_distribution(
        self, observations: Union[Tensor, Dict[str, Tensor]]
    ) -> T.distributions.Distribution:
        """Get the policy distribution given an observation"""
        self.model.eval()
        return self.model.predict_distribution(observations)

    @T.no_grad()
    def critic(
        self,
        observations: Union[Tensor, Dict[str, Tensor]],
        actions: Optional[Tensor] = None,
    ) -> T.Tensor:
        """Run the agent critic model"""
        self.model.eval()
        return self.model.predict_critic(observations, actions)

    def dump_log(self) -> None:
        """
        Write and reset the logger
        """
        self.logger.write_log(self.step)
        self.logger.reset_log()

    def step_env(self, observation: Observation, num_steps: int = 1) -> np.ndarray:
        """
        Step the agent in the environment

        :param observation: the starting observation to step from
        :param num_steps: how many steps to take
        :return: the final observation after all steps have been done
        """
        self.model.eval()
        for _ in range(num_steps):
            if self.render:
                self.env.render()
            action = self.action_explorer(self.model, observation, self.step)
            next_observation, reward, done, _ = self.env.step(action)
            self.buffer.add_trajectory(
                observation, action, reward, next_observation, done
            )
            self.logger.debug(
                f"{Trajectories(observation, action, reward, next_observation, done)}"
            )
            # Add reward to current episode log
            self.logger.add_reward(reward)
            observation = next_observation

            # If all environment episodes are done, reset and check if we should dump the log
            if self.logger.check_episode_done(done):
                observation = self.env.reset()
                if self.log_frequency[0] == FrequencyType.EPISODE:
                    if self.episode % self.log_frequency[1] == 0:
                        self.dump_log()
                self.episode += 1

            if self.log_frequency[0] == FrequencyType.STEP:
                if self.step % self.log_frequency[1] == 0:
                    self.dump_log()
            if self.callbacks is not None:
                if not all(
                    [callback.on_step(self.step) for callback in self.callbacks]
                ):
                    self.done = True
                    break
            self.step += 1
        return observation

    @abstractmethod
    def _fit(
        self, batch_size: int, actor_epochs: int = 1, critic_epochs: int = 1
    ) -> Log:
        """
        Train the agent in the environment

        :param batch_size: minibatch size to make a single gradient descent step on
        :param actor_epochs: how many times to update the actor network in each training step
        :param critic_epochs: how many times to update the critic network in each training step
        :return: a Log object with training diagnostic info
        """

    def fit(
        self,
        num_steps: int,
        batch_size: int,
        actor_epochs: int = 1,
        critic_epochs: int = 1,
        train_frequency: Tuple[str, int] = ("step", 1),
    ) -> None:
        """
        Train the agent in the environment

        :param num_steps: total number of environment steps to train over
        :param batch_size: minibatch size to make a single gradient descent step on
        :param actor_epochs: how many times to update the actor network in each training step
        :param critic_epochs: how many times to update the critic network in each training step
        :param train_frequency: the number of steps or episodes to run before running a training step.
            To run every n episodes, use `("episode", n)`.
            To run every n steps, use `("step", n)`.
        """
        train_frequency = (
            FrequencyType(train_frequency[0].lower()),
            train_frequency[1],
        )
        # We can pre-calculate how many training steps to run if train_frequency is in steps rather than episodes
        if train_frequency[0] == FrequencyType.STEP:
            num_steps = num_steps // train_frequency[1]

        observation = self.env.reset()
        for step in range(num_steps):
            # Always fill buffer with enough samples for first training step
            if step == 0:
                observation = self.step_env(
                    observation=observation, num_steps=batch_size
                )
            # Step for number of steps specified
            elif train_frequency[0] == FrequencyType.STEP:
                observation = self.step_env(
                    observation=observation, num_steps=train_frequency[1]
                )
            # Step for number of episodes specified
            elif train_frequency[0] == FrequencyType.EPISODE:
                start_episode = self.episode
                end_episode = start_episode + train_frequency[1]
                while self.episode != end_episode:
                    observation = self.step_env(observation=observation)
                if self.step >= num_steps:
                    break

            if self.done:
                break

            self.model.train()
            train_log = self._fit(
                batch_size=batch_size,
                actor_epochs=actor_epochs,
                critic_epochs=critic_epochs,
            )
            self.model.update_global()
            self.logger.add_train_log(train_log)



import os

import torch as T

from pearll.callbacks.base_callback import BaseCallback
from pearll.common.logging_ import Logger
from pearll.models.actor_critics import ActorCritic


class CheckpointCallback(BaseCallback):
    def __init__(
        self,
        logger: Logger,
        model: ActorCritic,
        save_freq: int,
        save_path: str,
        name_prefix: str = "agent",
    ) -> None:
        super().__init__(logger, model)
        self.save_freq = save_freq
        self.save_path = save_path
        self.name_prefix = name_prefix

        os.makedirs(save_path, exist_ok=True)

    def load(self, path: str):
        """Load the model"""
        path = path + ".pt"
        self.logger.info(f"Loading weights from {path}")
        try:
            self.model.load_state_dict(T.load(path))
        except FileNotFoundError:
            self.logger.info("File not found, assuming no model dict was to be loaded")

    def save(self, path: str):
        """Save the model"""
        path = path + ".pt"
        self.logger.info(f"Saving weights to {path}")
        T.save(self.model.state_dict(), path)

    def _on_step(self) -> bool:
        if self.n_calls % self.save_freq == 0:
            path = os.path.join(self.save_path, f"{self.name_prefix}_{self.step}_steps")
            self.save(path)
        return True



from pearll.models.actor_critics import (
    Actor,
    ActorCritic,
    Critic,
    Dummy,
    EpsilonGreedyActor,
)
from pearll.models.environment import ModelEnv

__all__ = [
    "Critic",
    "Actor",
    "Dummy",
    "ActorCritic",
    "EpsilonGreedyActor",
    "ModelEnv",
]



import warnings
from functools import partial
from typing import Callable, List, Optional, Type

import numpy as np
from gym.vector.vector_env import VectorEnv
from sklearn.preprocessing import scale

from pearll.agents.base_agents import BaseAgent
from pearll.buffers import RolloutBuffer
from pearll.buffers.base_buffer import BaseBuffer
from pearll.callbacks.base_callback import BaseCallback
from pearll.common.type_aliases import Log
from pearll.common.utils import filter_rewards
from pearll.explorers.base_explorer import BaseExplorer
from pearll.models import ActorCritic, Dummy
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    MiscellaneousSettings,
    MutationSettings,
    PopulationSettings,
    Settings,
)
from pearll.updaters.evolution import BaseEvolutionUpdater, NoisyGradientAscent

warnings.filterwarnings("ignore", category=UserWarning)


def default_model(env: VectorEnv):
    """
    Returns a default model for the given environment.
    """
    actor = Dummy(space=env.single_action_space)
    critic = Dummy(space=env.single_action_space)

    return ActorCritic(
        actor=actor,
        critic=critic,
        population_settings=PopulationSettings(
            actor_population_size=env.num_envs, actor_distribution="normal"
        ),
    )


class AdamES(BaseAgent):
    """
    AdamES is a variant of ES that uses the Adam optimizer.

    :param env: the gym-like environment to be used, should be a VectorEnv
    :param model: the neural network model
    :param updater_class: the updater class to be used
    :param mutation_operator: the mutation operator to be used
    :param mutation_settings: the mutation settings to be used
    :param learning_rate: the learning rate to be used
    :param momentum_weight: the adam momentum weight to be used
    :param damping_weight: the adam damping weight to be used
    :param learning_rate: the learning rate
    :param buffer_class: the buffer class for storing and sampling trajectories
    :param buffer_settings: settings for the buffer
    :param action_explorer_class: the explorer class for random search at beginning of training and
        adding noise to actions
    :param explorer settings: settings for the action explorer
    :param callbacks: an optional list of callbacks (e.g. if you want to save the model)
    :param callback_settings: settings for callbacks
    :param logger_settings: settings for the logger
    :param misc_settings: settings for miscellaneous parameters
    """

    def __init__(
        self,
        env: VectorEnv,
        model: Optional[ActorCritic] = None,
        updater_class: Type[BaseEvolutionUpdater] = NoisyGradientAscent,
        mutation_operator: Optional[Callable] = None,
        mutation_settings: MutationSettings = MutationSettings(mutation_std=0.5),
        learning_rate: float = 1,
        momentum_weight: float = 0.9,
        damping_weight: float = 0.999,
        buffer_class: Type[BaseBuffer] = RolloutBuffer,
        buffer_settings: BufferSettings = BufferSettings(),
        action_explorer_class: Type[BaseExplorer] = BaseExplorer,
        explorer_settings: ExplorerSettings = ExplorerSettings(start_steps=0),
        callbacks: Optional[List[Type[BaseCallback]]] = None,
        callback_settings: Optional[List[Settings]] = None,
        logger_settings: LoggerSettings = LoggerSettings(),
        misc_settings: MiscellaneousSettings = MiscellaneousSettings(),
    ) -> None:
        model = model if model is not None else default_model(env)
        super().__init__(
            env=env,
            model=model,
            action_explorer_class=action_explorer_class,
            explorer_settings=explorer_settings,
            buffer_class=buffer_class,
            buffer_settings=buffer_settings,
            logger_settings=logger_settings,
            callbacks=callbacks,
            callback_settings=callback_settings,
            misc_settings=misc_settings,
        )

        self.learning_rate = learning_rate
        self.momentum_weight = momentum_weight
        self.damping_weight = damping_weight
        self.updater = updater_class(model=self.model)
        self.mutation_operator = (
            None
            if mutation_operator is None
            else partial(mutation_operator, **mutation_settings.filter_none())
        )
        self.m = 0
        self.v = 0
        self.adam_step = 1

    def _adam(self, grad: np.floating) -> np.floating:
        """Adam optimizer update"""
        self.m = (1 - self.momentum_weight) * grad + self.momentum_weight * self.m
        self.v = (1 - self.damping_weight) * (
            grad * grad
        ) + self.damping_weight * self.v
        m_adj = self.m / (1 - (self.momentum_weight ** self.adam_step))
        v_adj = self.v / (1 - (self.damping_weight ** self.adam_step))
        self.adam_step += 1
        return m_adj / (np.sqrt(v_adj) + 1e-8)

    def _fit(
        self, batch_size: int, actor_epochs: int = 1, critic_epochs: int = 1
    ) -> Log:
        divergences = np.zeros(actor_epochs)
        entropies = np.zeros(actor_epochs)

        trajectories = self.buffer.all(dtype="numpy")
        rewards = trajectories.rewards.squeeze()
        rewards = filter_rewards(rewards, trajectories.dones.squeeze())
        if rewards.ndim > 1:
            rewards = rewards.sum(axis=-1)
        scaled_rewards = scale(rewards)
        grad_approx = np.dot(self.updater.normal_dist.T, scaled_rewards) / (
            np.mean(self.updater.std) * self.env.num_envs
        )
        optimization_direction = self._adam(grad_approx)
        for i in range(actor_epochs):
            log = self.updater(
                learning_rate=self.learning_rate,
                optimization_direction=optimization_direction,
                mutation_operator=self.mutation_operator,
            )
            divergences[i] = log.divergence
            entropies[i] = log.entropy
        self.buffer.reset()

        return Log(divergence=divergences.sum(), entropy=entropies.mean())



import warnings
from abc import ABC, abstractmethod
from typing import Union

import numpy as np
import psutil
import torch as T
from gym import Env
from gym.vector import VectorEnv

from pearll.common.enumerations import TrajectoryType
from pearll.common.type_aliases import Observation, Trajectories
from pearll.common.utils import get_device, get_space_shape, to_torch


class BaseBuffer(ABC):
    """
    the base buffer class which handles sample collection and processing.

    :param env: the environment
    :param buffer_size: max number of elements in the buffer
    :param device: if return torch tensors on sampling, the device to attach to
    """

    def __init__(
        self,
        env: Env,
        buffer_size: int,
        device: Union[str, T.device] = "auto",
    ) -> None:
        self.env = env
        self.buffer_size = buffer_size
        self.device = get_device(device)
        self.full = False
        self.pos = 0

        self.num_envs = env.num_envs if isinstance(env, VectorEnv) else 1

        # If only 1 environment, don't need the num_envs axis
        if self.num_envs > 1:
            self.batch_shape = (buffer_size, self.num_envs)
        else:
            self.batch_shape = (buffer_size,)

        self.obs_shape = get_space_shape(env.observation_space)
        self.action_shape = get_space_shape(env.action_space)

        self.observations = np.zeros(
            (self.buffer_size,) + self.obs_shape,
            dtype=env.observation_space.dtype,
        )
        self.actions = np.zeros(
            (self.buffer_size,) + self.action_shape,
            dtype=env.action_space.dtype,
        )
        # Use 3 dims for easier calculations without having to think about broadcasting
        self.rewards = np.zeros(self.batch_shape + (1,), dtype=np.float32)
        self.dones = np.zeros(self.batch_shape + (1,), dtype=np.float32)

    @staticmethod
    def _check_system_memory(*buffers) -> None:
        """Check that the replay buffer can fit into memory"""
        mem_available = psutil.virtual_memory().available
        total_memory_usage = sum([buffer.nbytes for buffer in buffers])

        if total_memory_usage > mem_available:
            # Convert to GB
            total_memory_usage /= 1e9
            mem_available /= 1e9
            warnings.warn(
                "This system does not have enough memory to store the complete "
                f"replay buffer: {total_memory_usage:.2f}GB > {mem_available:.2f}GB"
            )

    def _flatten_env_axis(
        self,
        data: np.ndarray,
    ) -> np.ndarray:
        """
        Flatten the n_env axis of the input arrays to get samples of (batch_size, ...)

        :param data: the data to flatten (batch_size, num_envs, ...)
        :return: the flattened data (batch_size, ...)
        """
        if self.num_envs > 1:
            batch_size = data.shape[0] // self.num_envs
            data = data[-batch_size:].reshape(
                batch_size * self.num_envs, *data.shape[2:]
            )

        return data

    def _transform_samples(
        self,
        flatten_env: bool,
        dtype: Union[str, TrajectoryType],
        observations: np.ndarray,
        actions: np.ndarray,
        rewards: np.ndarray,
        next_observations: np.ndarray,
        dones: np.ndarray,
    ) -> Trajectories:
        """
        Handle post-processing of sampled trajectories:
        1. Transform to torch tensor if specified
        2. Flatten the n_env axis if specified

        :param flatten_env: whether to flatten the num_envs axis
        :param dtype: the data type to return (torch or numpy)
        :param observations: the observations
        :param actions: the actions
        :param rewards: the rewards
        :param next_observations: the next observations
        :param dones: the done flags
        :return: the final transformed trajectories
        """
        if isinstance(dtype, str):
            dtype = TrajectoryType(dtype.lower())

        if flatten_env:
            observations = self._flatten_env_axis(observations)
            actions = self._flatten_env_axis(actions)
            rewards = self._flatten_env_axis(rewards)
            next_observations = self._flatten_env_axis(next_observations)
            dones = self._flatten_env_axis(dones)
        elif self.num_envs > 1:
            observations = observations.swapaxes(0, 1)
            actions = actions.swapaxes(0, 1)
            rewards = rewards.swapaxes(0, 1)
            next_observations = next_observations.swapaxes(0, 1)
            dones = dones.swapaxes(0, 1)

        # return torch tensors instead of numpy arrays
        if dtype == TrajectoryType.TORCH:
            observations = to_torch(observations, device=self.device)
            actions = to_torch(actions, device=self.device)
            rewards = to_torch(rewards, device=self.device)
            next_observations = to_torch(next_observations, device=self.device)
            dones = to_torch(dones, device=self.device)

        return Trajectories(
            observations=observations,
            actions=actions,
            rewards=rewards,
            next_observations=next_observations,
            dones=dones,
        )

    @abstractmethod
    def reset(self) -> None:
        """Reset the buffer"""

        self.pos = 0
        self.full = False

        self.observations = np.zeros(
            (self.buffer_size,) + self.obs_shape,
            dtype=self.env.observation_space.dtype,
        )
        self.actions = np.zeros(
            (self.buffer_size,) + self.action_shape,
            dtype=self.env.action_space.dtype,
        )
        self.rewards = np.zeros(self.batch_shape + (1,), dtype=np.float32)
        self.dones = np.zeros(self.batch_shape + (1,), dtype=np.float32)

    @abstractmethod
    def add_trajectory(
        self,
        observation: Observation,
        action: Union[np.ndarray, int],
        reward: Union[float, np.ndarray],
        next_observation: Observation,
        done: Union[bool, np.ndarray],
    ) -> None:
        """
        Add a trajectory to the buffer

        :param observation: the observation
        :param action: the action taken
        :param reward: the reward received
        :param next_observation: the next observation collected
        :param done: the trajectory done flag
        """

    def add_batch_trajectories(
        self,
        observations: Observation,
        actions: np.ndarray,
        rewards: np.ndarray,
        next_observations: Observation,
        dones: np.ndarray,
    ) -> None:
        """
        Add a batch of trajectories to the buffer

        :param observations: the observations
        :param action: the actions taken
        :param reward: the rewards received
        :param next_observations: the next observations collected
        :param done: the trajectory done flags
        """
        for data in zip(observations, actions, rewards, next_observations, dones):
            self.add_trajectory(*data)

    @abstractmethod
    def sample(
        self,
        batch_size: int,
        flatten_env: bool = False,
        dtype: Union[str, TrajectoryType] = "torch",
    ) -> Trajectories:
        """
        Sample a batch of trajectories

        :param batch_size: the batch size
        :param flatten_env: useful for multiple environments, whether to sample with the num_envs axis
        :param dtype: whether to return the trajectories as "numpy" or "torch", default torch
        :return: the sampled trajectories
        """

    @abstractmethod
    def last(
        self,
        batch_size: int,
        flatten_env: bool = False,
        dtype: Union[str, TrajectoryType] = "torch",
    ) -> Trajectories:
        """
        Get the most recent batch of trajectories stored

        :param batch_size: the batch size
        :param flatten_env: useful for multiple environments, whether to sample with the num_envs axis
        :param dtype: whether to return the trajectories as "numpy" or "torch", default torch
        :return: the most recent trajectories
        """

    @abstractmethod
    def all(
        self, flatten_env: bool = False, dtype: Union[str, TrajectoryType] = "torch"
    ) -> Trajectories:
        """
        Get all stored trajectories

        :param flatten_env: useful for multiple environments, whether to sample with the num_envs axis
        :param dtype: whether to return the trajectories as "numpy" or "torch", default torch
        :return: stored trajectories
        """



from typing import Any, Optional, Tuple

from gym import Env, Space

from pearll.common import utils
from pearll.common.type_aliases import DoneFunc, ObservationFunc, RewardFunc


class ModelEnv(Env):
    """
    Model the environment, M(S, A) -> R, S'

    :param reward_fn: reward_fn(S, A) -> R
    :param observation_fn: observation_fn(S, A) -> S'
    :param done_fn: optional done_fn(S, A) -> done
    :param reset_space: observation space sampled to reset the environment
    """

    def __init__(
        self,
        reward_fn: RewardFunc,
        observation_fn: ObservationFunc,
        done_fn: Optional[DoneFunc],
        reset_space: Space,
    ) -> None:
        self.reward_fn = reward_fn
        self.observation_fn = observation_fn
        self.done_fn = done_fn
        self.reset_space = reset_space

    def step(self, observation: Any, action: Any) -> Tuple[Any, float, bool, dict]:
        """
        Step in the environment

        :param observation: observation
        :param action: action
        :return: observation, reward, done, info
        """
        next_observation = self.observation_fn(observation, action)
        reward = self.reward_fn(observation, action)
        done = False if self.done_fn is None else self.done_fn(observation, action)
        return utils.to_numpy(next_observation), reward, done, {}

    def reset(self) -> Any:
        """
        Reset the environment

        :return: observation
        """
        return self.reset_space.sample()



from argparse import ArgumentParser
from typing import Any, Dict, Optional, OrderedDict, Tuple, Union

import gym
import numpy as np
import torch as T

from pearll.agents import A2C, CEM_RL, DDPG, DQN, ES, GA, PPO, AdamES, DynaQ
from pearll.buffers import HERBuffer
from pearll.common.utils import get_space_shape
from pearll.models import ActorCritic, Critic, Dummy, EpsilonGreedyActor
from pearll.models.actor_critics import Actor, Model
from pearll.models.encoders import DictEncoder, IdentityEncoder
from pearll.models.environment import ModelEnv
from pearll.models.heads import BoxHead, CategoricalHead, DiscreteHead, DiscreteQHead
from pearll.models.torsos import MLP
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    OptimizerSettings,
    PopulationSettings,
)
from pearll.updaters.environment import DeepRegression


class Sphere(gym.Env):
    """
    Sphere(2) function for testing ES agent.
    """

    def __init__(self):
        self.action_space = gym.spaces.Box(low=-100, high=100, shape=(2,))
        self.observation_space = gym.spaces.Discrete(1)

    def step(self, action):
        return 0, -(action[0] ** 2 + action[1] ** 2), False, {}

    def reset(self):
        return 0


def dqn_demo():
    env = gym.make("CartPole-v0")
    agent = DQN(
        env=env,
        model=None,
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/DQN-demo", verbose=True
        ),
        explorer_settings=ExplorerSettings(start_steps=1000),
    )
    agent.fit(
        num_steps=20000, batch_size=32, critic_epochs=16, train_frequency=("episode", 1)
    )


def dqn_parallel_demo():
    env = gym.vector.make("CartPole-v0", 2, asynchronous=True)
    encoder = IdentityEncoder()
    torso = MLP(layer_sizes=[4, 64, 32], activation_fn=T.nn.ReLU)
    head = DiscreteQHead(input_shape=32, output_shape=2)

    actor = EpsilonGreedyActor(
        critic_encoder=encoder, critic_torso=torso, critic_head=head
    )
    critic = Critic(encoder=encoder, torso=torso, head=head, create_target=True)
    model = ActorCritic(
        actor=actor,
        critic=critic,
        population_settings=PopulationSettings(
            actor_population_size=2,
            critic_population_size=2,
            actor_distribution="normal",
            critic_distribution="normal",
        ),
    )

    agent = DQN(
        env=env,
        model=model,
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/DQN-parallel-demo", log_frequency=("episode", 1)
        ),
        explorer_settings=ExplorerSettings(start_steps=1000),
    )
    agent.fit(
        num_steps=20000,
        batch_size=32,
        critic_epochs=16,
        train_frequency=("episode", 1),
    )


def ddpg_demo():
    env = gym.make("Pendulum-v0")
    agent = DDPG(
        env=env,
        model=None,
        logger_settings=LoggerSettings(tensorboard_log_path="runs/DDPG-demo"),
        explorer_settings=ExplorerSettings(start_steps=1000, scale=0.1),
    )
    agent.fit(
        num_steps=100000,
        batch_size=64,
        critic_epochs=1,
        actor_epochs=1,
        train_frequency=("step", 1),
    )


def es_demo():
    POPULATION_SIZE = 10
    env = gym.vector.SyncVectorEnv([lambda: Sphere() for _ in range(POPULATION_SIZE)])
    actor = Dummy(space=env.single_action_space, state=np.array([10, 10]))
    critic = Dummy(space=env.single_action_space, state=np.array([10, 10]))

    model = ActorCritic(
        actor=actor,
        critic=critic,
        population_settings=PopulationSettings(
            actor_population_size=env.num_envs, actor_distribution="normal"
        ),
    )

    agent = ES(
        env=env,
        model=model,
        learning_rate=1,
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/ES-demo", log_frequency=("step", 1)
        ),
    )
    agent.fit(num_steps=20, batch_size=1)


def es_deep_demo():
    POPULATION_SIZE = 20
    env = gym.vector.make("CartPole-v0", POPULATION_SIZE, asynchronous=True)
    observation_shape = get_space_shape(env.single_observation_space)
    encoder = IdentityEncoder()
    torso = MLP(layer_sizes=[observation_shape[0], 20, 10], activation_fn=T.nn.ReLU)
    head = CategoricalHead(
        input_shape=10, action_size=env.single_action_space.n, activation_fn=T.nn.Tanh
    )
    actor = Actor(encoder=encoder, torso=torso, head=head)
    critic = Dummy(space=env.single_action_space)
    model = ActorCritic(
        actor=actor,
        critic=critic,
        population_settings=PopulationSettings(
            actor_population_size=POPULATION_SIZE,
            actor_distribution="normal",
            actor_std=0.01,
        ),
    )

    agent = ES(
        env=env,
        model=model,
        learning_rate=0.001,
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/DeepES-demo", log_frequency=("episode", 1)
        ),
    )
    agent.fit(
        num_steps=100000, batch_size=0, actor_epochs=1, train_frequency=("episode", 1)
    )


def ga_demo():
    class Mastermind(gym.Env):
        """
        Mastermind game for testing GA agent.
        """

        def __init__(self):
            self.action_space = gym.spaces.MultiDiscrete([4, 4, 4, 4])
            self.observation_space = gym.spaces.Discrete(1)
            self.master = [0, 1, 2, 3]

        def step(self, action):
            p1 = 0
            p2 = 0
            p1_map = [0] * 4
            p2_map = [0] * 4
            Mp2_map = [0] * 4

            for i, colour in enumerate(action):
                if colour == self.master[i]:
                    p1_map[i] = 1
                for el in self.master:
                    if colour == el:
                        p2_map[i] = 1
                for i, el in enumerate(self.master):
                    for colour in action:
                        if colour == el:
                            Mp2_map[i] = 1

            Mp2_map = [m - n for m, n in zip(Mp2_map, p1_map)]
            p2_map = [m - n for m, n in zip(p2_map, p1_map)]

            p1 = sum(p1_map)
            p2 = min(sum(p2_map), sum(Mp2_map))

            return 0, p1 + 0.5 * p2, False, {}

        def reset(self):
            return 0

    POPULATION_SIZE = 10
    env = gym.vector.SyncVectorEnv(
        [lambda: Mastermind() for _ in range(POPULATION_SIZE)]
    )

    agent = GA(
        env=env,
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/GA-demo", log_frequency=("step", 1)
        ),
    )
    agent.fit(num_steps=50, batch_size=1)


def her_demo():
    class BitFlippingEnv(gym.GoalEnv):
        """
        Simple bit flipping env, useful to test HER.
        The goal is to flip all the bits to get a vector of ones.
        In the continuous variant, if the ith action component has a value > 0,
        then the ith bit will be flipped.
        :param n_bits: Number of bits to flip
        :param continuous: Whether to use the continuous actions version or not,
            by default, it uses the discrete one
        :param max_steps: Max number of steps, by default, equal to n_bits
        :param discrete_obs_space: Whether to use the discrete observation
            version or not, by default, it uses the ``MultiBinary`` one
        :param image_obs_space: Use image as input instead of the ``MultiBinary`` one.
        :param channel_first: Whether to use channel-first or last image.
        """

        spec = gym.envs.registration.EnvSpec("BitFlippingEnv-v0")

        def __init__(
            self,
            n_bits: int = 10,
            continuous: bool = False,
            max_steps: Optional[int] = None,
            discrete_obs_space: bool = False,
            image_obs_space: bool = False,
            channel_first: bool = True,
        ):
            super(BitFlippingEnv, self).__init__()
            # Shape of the observation when using image space
            self.image_shape = (1, 36, 36) if channel_first else (36, 36, 1)
            # The achieved goal is determined by the current state
            # here, it is a special where they are equal
            if discrete_obs_space:
                # In the discrete case, the agent act on the binary
                # representation of the observation
                self.observation_space = gym.spaces.Dict(
                    {
                        "observation": gym.spaces.Discrete(2 ** n_bits),
                        "achieved_goal": gym.spaces.Discrete(2 ** n_bits),
                        "desired_goal": gym.spaces.Discrete(2 ** n_bits),
                    }
                )
            elif image_obs_space:
                # When using image as input,
                # one image contains the bits 0 -> 0, 1 -> 255
                # and the rest is filled with zeros
                self.observation_space = gym.spaces.Dict(
                    {
                        "observation": gym.spaces.Box(
                            low=0,
                            high=255,
                            shape=self.image_shape,
                            dtype=np.uint8,
                        ),
                        "achieved_goal": gym.spaces.Box(
                            low=0,
                            high=255,
                            shape=self.image_shape,
                            dtype=np.uint8,
                        ),
                        "desired_goal": gym.spaces.Box(
                            low=0,
                            high=255,
                            shape=self.image_shape,
                            dtype=np.uint8,
                        ),
                    }
                )
            else:
                self.observation_space = gym.spaces.Dict(
                    {
                        "observation": gym.spaces.MultiBinary(n_bits),
                        "achieved_goal": gym.spaces.MultiBinary(n_bits),
                        "desired_goal": gym.spaces.MultiBinary(n_bits),
                    }
                )

            self.obs_space = gym.spaces.MultiBinary(n_bits)

            if continuous:
                self.action_space = gym.spaces.Box(
                    -1, 1, shape=(n_bits,), dtype=np.float32
                )
            else:
                self.action_space = gym.spaces.Discrete(n_bits)
            self.continuous = continuous
            self.discrete_obs_space = discrete_obs_space
            self.image_obs_space = image_obs_space
            self.state = None
            self.desired_goal = np.ones((n_bits,))
            if max_steps is None:
                max_steps = n_bits
            self.max_steps = max_steps
            self.current_step = 0

        def seed(self, seed: int) -> None:
            self.obs_space.seed(seed)

        def convert_if_needed(self, state: np.ndarray) -> Union[int, np.ndarray]:
            """
            Convert to discrete space if needed.
            :param state:
            :return:
            """
            if self.discrete_obs_space:
                # The internal state is the binary representation of the
                # observed one
                return int(sum([state[i] * 2 ** i for i in range(len(state))]))

            if self.image_obs_space:
                size = np.prod(self.image_shape)
                image = np.concatenate(
                    (state * 255, np.zeros(size - len(state), dtype=np.uint8))
                )
                return image.reshape(self.image_shape).astype(np.uint8)
            return state

        def convert_to_bit_vector(
            self, state: Union[int, np.ndarray], batch_size: int
        ) -> np.ndarray:
            """
            Convert to bit vector if needed.
            :param state:
            :param batch_size:
            :return:
            """
            # Convert back to bit vector
            if isinstance(state, int):
                state = np.array(state).reshape(batch_size, -1)
                # Convert to binary representation
                state = (
                    ((state[:, :] & (1 << np.arange(len(self.state))))) > 0
                ).astype(int)
            elif self.image_obs_space:
                state = state.reshape(batch_size, -1)[:, : len(self.state)] / 255
            else:
                state = np.array(state).reshape(batch_size, -1)

            return state

        def _get_obs(self) -> Dict[str, Union[int, np.ndarray]]:
            """
            Helper to create the observation.
            :return: The current observation.
            """
            return OrderedDict(
                [
                    ("observation", self.convert_if_needed(self.state.copy())),
                    ("achieved_goal", self.convert_if_needed(self.state.copy())),
                    ("desired_goal", self.convert_if_needed(self.desired_goal.copy())),
                ]
            )

        def reset(self) -> Dict[str, Union[int, np.ndarray]]:
            self.current_step = 0
            self.state = self.obs_space.sample()
            return self._get_obs()

        def step(self, action: Union[np.ndarray, int]) -> Tuple:
            if self.continuous:
                self.state[action > 0] = 1 - self.state[action > 0]
            else:
                self.state[action] = 1 - self.state[action]
            obs = self._get_obs()
            reward = float(
                self.compute_reward(obs["achieved_goal"], obs["desired_goal"], None)
            )
            done = reward == 0
            self.current_step += 1
            # Episode terminate when we reached the goal or the max number of steps
            info = {"is_success": done}
            done = done or self.current_step >= self.max_steps
            return obs, reward, done, info

        def compute_reward(
            self,
            achieved_goal: Union[int, np.ndarray],
            desired_goal: Union[int, np.ndarray],
            _info: Optional[Dict[str, Any]],
        ) -> np.float32:
            # As we are using a vectorized version, we need to keep track of the `batch_size`
            if isinstance(achieved_goal, int):
                batch_size = 1
            elif self.image_obs_space:
                batch_size = (
                    achieved_goal.shape[0] if len(achieved_goal.shape) > 3 else 1
                )
            else:
                batch_size = (
                    achieved_goal.shape[0] if len(achieved_goal.shape) > 1 else 1
                )

            desired_goal = self.convert_to_bit_vector(desired_goal, batch_size)
            achieved_goal = self.convert_to_bit_vector(achieved_goal, batch_size)

            # Deceptive reward: it is positive only when the goal is achieved
            # Here we are using a vectorized version
            distance = np.linalg.norm(achieved_goal - desired_goal, axis=-1)
            return -(distance > 0).astype(np.float32)

        def render(self, mode: str = "human") -> Optional[np.ndarray]:
            if mode == "rgb_array":
                return self.state.copy()
            print(self.state)

        def close(self) -> None:
            pass

    env = BitFlippingEnv()

    action_size = env.action_space.n
    observation_size = 20

    encoder = DictEncoder()
    torso = MLP(layer_sizes=[observation_size, 256], activation_fn=T.nn.ReLU)
    head = DiscreteQHead(input_shape=256, output_shape=action_size)

    actor = EpsilonGreedyActor(
        critic_encoder=encoder, critic_torso=torso, critic_head=head
    )
    critic = Critic(encoder=encoder, torso=torso, head=head, create_target=True)
    model = ActorCritic(actor=actor, critic=critic)

    agent = DQN(
        env=env,
        model=model,
        buffer_class=HERBuffer,
        buffer_settings=BufferSettings(),
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/HER-demo", verbose=True
        ),
        explorer_settings=ExplorerSettings(start_steps=0),
    )

    agent.fit(
        num_steps=50000,
        batch_size=128,
        critic_epochs=40,
        train_frequency=("episode", 16),
    )


def a2c_demo():
    env = gym.make("CartPole-v0")
    batch_size = 5

    agent = A2C(
        env=env,
        buffer_settings=BufferSettings(buffer_size=batch_size),
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/A2C-demo", verbose=True
        ),
        entropy_coeff=0,
    )

    agent.fit(
        num_steps=200000,
        batch_size=batch_size,
        critic_epochs=1,
        train_frequency=("step", batch_size),
    )


def ppo_demo():
    env = gym.make("CartPole-v0")
    batch_size = 5
    epochs = 1

    agent = PPO(
        env=env,
        buffer_settings=BufferSettings(buffer_size=batch_size),
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/PPO-demo", verbose=True
        ),
        entropy_coeff=0,
    )

    agent.fit(
        num_steps=200000,
        batch_size=batch_size,
        actor_epochs=epochs,
        critic_epochs=epochs,
        train_frequency=("step", batch_size),
    )


def cem_ddpg_demo():
    POPULATION_SIZE = 20
    env = gym.vector.make("Pendulum-v0", POPULATION_SIZE, asynchronous=True)
    eval_env = gym.vector.make("Pendulum-v0", POPULATION_SIZE, asynchronous=True)

    agent = CEM_RL(
        env=env,
        eval_env=eval_env,
        model=None,
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/CEM-DDPG-demo", log_frequency=("episode", 1)
        ),
    )

    agent.fit(
        num_steps=50000,
        batch_size=64,
        actor_epochs=1,
        critic_epochs=8,
        train_frequency=("step", 100),
    )


def adames_demo():
    POPULATION_SIZE = 10
    env = gym.vector.SyncVectorEnv([lambda: Sphere() for _ in range(POPULATION_SIZE)])
    actor = Dummy(space=env.single_action_space, state=np.array([10, 10]))
    critic = Dummy(space=env.single_action_space, state=np.array([10, 10]))

    model = ActorCritic(
        actor=actor,
        critic=critic,
        population_settings=PopulationSettings(
            actor_population_size=env.num_envs, actor_distribution="normal"
        ),
    )

    agent = AdamES(
        env=env,
        model=model,
        learning_rate=1,
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/AdamES-demo", log_frequency=("step", 1)
        ),
    )
    agent.fit(num_steps=50, batch_size=1)


def dynaq_demo():
    env = gym.make("CartPole-v0")

    encoder = IdentityEncoder()
    torso = MLP(layer_sizes=[4, 64, 32], activation_fn=T.nn.ReLU)
    head = DiscreteQHead(input_shape=32, output_shape=2)

    actor = EpsilonGreedyActor(
        critic_encoder=encoder, critic_torso=torso, critic_head=head
    )
    critic = Critic(encoder=encoder, torso=torso, head=head, create_target=True)
    agent_model = ActorCritic(actor=actor, critic=critic)

    class CartPoleObsActivation(T.nn.Module):
        def __init__(self) -> None:
            super().__init__()
            self.tanh = T.nn.Tanh()

        def forward(self, input: T.Tensor) -> T.Tensor:
            return 4 * self.tanh(input)

    encoder = IdentityEncoder()

    obs_torso = MLP(layer_sizes=[5, 64, 32], activation_fn=T.nn.ReLU)
    obs_head = BoxHead(
        input_shape=32, space_shape=4, activation_fn=CartPoleObsActivation
    )
    obs_model = Model(encoder, obs_torso, obs_head)

    reward_torso = MLP(layer_sizes=[5, 64, 32], activation_fn=T.nn.ReLU)
    reward_head = DiscreteHead(input_shape=32, space_size=2)
    reward_model = Model(encoder, reward_torso, reward_head)

    done_torso = MLP(layer_sizes=[5, 64, 32], activation_fn=T.nn.ReLU)
    done_head = DiscreteHead(input_shape=32, space_size=2, dtype="bool")
    done_model = Model(encoder, done_torso, done_head)

    reset_space = gym.spaces.Box(low=-0.05, high=0.05, shape=(4,))
    env_model = ModelEnv(
        reward_fn=reward_model,
        observation_fn=obs_model,
        done_fn=done_model,
        reset_space=reset_space,
    )

    agent = DynaQ(
        env=env,
        agent_model=agent_model,
        env_model=env_model,
        done_updater_class=DeepRegression,
        reward_optimizer_settings=OptimizerSettings(loss_class=T.nn.BCELoss()),
        logger_settings=LoggerSettings(
            tensorboard_log_path="runs/DynaQ-demo", verbose=True
        ),
    )
    agent.fit(
        env_steps=5000,
        plan_steps=2000,
        env_batch_size=16,
        plan_batch_size=32,
        critic_epochs=16,
        env_epochs=32,
        env_train_frequency=("episode", 1),
        plan_train_frequency=("step", 200),
        no_model_steps=0,
    )


if __name__ == "__main__":
    agents = [
        "a2c",
        "ppo",
        "cem-ddpg",
        "adames",
        "dynaq",
        "ddpg",
        "dqn",
        "es",
        "es-deep",
        "ga",
        "her",
    ]
    parser = ArgumentParser(description="pearll demo with preloaded hyperparameters")
    parser.add_argument("--agent", help="Agent to demo")
    kwargs = parser.parse_args()

    if kwargs.agent.lower() == "dqn":
        dqn_demo()
    elif kwargs.agent.lower() == "dqn-parallel":
        dqn_parallel_demo()
    elif kwargs.agent.lower() == "ddpg":
        ddpg_demo()
    elif kwargs.agent.lower() == "es":
        es_demo()
    elif kwargs.agent.lower() == "es-deep":
        es_deep_demo()
    elif kwargs.agent.lower() == "ga":
        ga_demo()
    elif kwargs.agent.lower() == "her":
        her_demo()
    elif kwargs.agent.lower() == "a2c":
        a2c_demo()
    elif kwargs.agent.lower() == "ppo":
        ppo_demo()
    elif kwargs.agent.lower() == "cem-ddpg":
        cem_ddpg_demo()
    elif kwargs.agent.lower() == "adames":
        adames_demo()
    elif kwargs.agent.lower() == "dynaq":
        dynaq_demo()
    else:
        raise ValueError(
            f"Agent {kwargs.agent} not found, valid arguments are {agents}"
        )



from abc import ABC, abstractmethod
from typing import Iterator, Type, Union

import torch as T
from torch.distributions import kl_divergence
from torch.nn.parameter import Parameter

from pearll.common.type_aliases import UpdaterLog
from pearll.models.actor_critics import Actor, ActorCritic


class BaseActorUpdater(ABC):
    """
    The base actor updater class with pre-defined methods for derived classes

    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        self.optimizer_class = optimizer_class
        self.max_grad = max_grad

    def _get_model_parameters(
        self, model: Union[Actor, ActorCritic]
    ) -> Iterator[Parameter]:
        """Get the actor model parameters"""
        if isinstance(model, Actor):
            return model.parameters()
        else:
            params = []
            for actor in model.actors:
                params.extend(actor.model.parameters())
            return params

    def run_optimizer(
        self,
        optimizer: T.optim.Optimizer,
        loss: T.Tensor,
        actor_parameters: Iterator[Parameter],
    ) -> None:
        """Run an optimization step"""
        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        if self.max_grad > 0:
            T.nn.utils.clip_grad_norm_(actor_parameters, self.max_grad)
        optimizer.step()

    @abstractmethod
    def __call__(self, model: Union[Actor, ActorCritic]) -> UpdaterLog:
        """Run an optimization step"""


class PolicyGradient(BaseActorUpdater):
    """
    Vanilla policy gradient with entropy regulation: https://spinningup.openai.com/en/latest/algorithms/vpg.html
    loss = -E[A(state,action) * log(policy(action|state)) + entropy_coeff * entropy(policy)]

    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        super().__init__(optimizer_class=optimizer_class, max_grad=max_grad)

    def __call__(
        self,
        model: Union[ActorCritic, Actor],
        observations: T.Tensor,
        actions: T.Tensor,
        advantages: T.Tensor,
        learning_rate: float = 1e-3,
        entropy_coeff: float = 0.01,
    ) -> UpdaterLog:
        """
        Perform an optimization step

        :param model: the model on which the optimization should be run
        :param observations:
        :param actions:
        :param advantages:
        :param learning_rate: the learning rate for the optimizer algorithm
        :param entropy_coeff: entropy regulation coefficient
        """
        actor_parameters = self._get_model_parameters(model)
        optimizer = self.optimizer_class(actor_parameters, lr=learning_rate)
        old_distributions = model.action_distribution(observations)
        log_probs = old_distributions.log_prob(actions).sum(dim=-1)
        entropy = old_distributions.entropy().mean()

        batch_loss = -(advantages * log_probs).mean()
        entropy_loss = -entropy_coeff * entropy

        loss = batch_loss + entropy_loss

        self.run_optimizer(optimizer, loss, actor_parameters)

        new_distributions = model.action_distribution(observations)
        loss = loss.detach()
        entropy = entropy.detach()
        kl = kl_divergence(new_distributions, old_distributions).mean()

        return UpdaterLog(
            loss=loss.item(), divergence=kl.item(), entropy=entropy.item()
        )


class ProximalPolicyClip(BaseActorUpdater):
    """
    PPO-Clip algorithm with entropy regularization: https://spinningup.openai.com/en/latest/algorithms/ppo.html
    loss = E[min(r)]

    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        super().__init__(optimizer_class=optimizer_class, max_grad=max_grad)

    def __call__(
        self,
        model: Union[ActorCritic, Actor],
        observations: T.Tensor,
        actions: T.Tensor,
        advantages: T.Tensor,
        old_log_probs: T.Tensor,
        learning_rate: float = 1e-3,
        ratio_clip: float = 0.2,
        entropy_coeff: float = 0.01,
    ) -> UpdaterLog:
        """
        Perform an optimization step

        :param model: the model on which the optimization should be run
        :param observations:
        :param actions:
        :param advantages:
        :param old_log_probs:
        :param learning_rate: the learning rate for the optimizer algorithm
        :param ratio_clip: the clipping factor for the clipped loss
        :param entropy_coeff: entropy regulation coefficient
        """
        actor_parameters = self._get_model_parameters(model)
        optimizer = self.optimizer_class(actor_parameters, lr=learning_rate)
        old_distributions = model.action_distribution(observations)
        log_probs = old_distributions.log_prob(actions).sum(dim=-1)
        entropy = old_distributions.entropy().mean()

        ratios = (log_probs - old_log_probs).exp()

        raw_loss = ratios * advantages
        clipped_loss = T.clamp(ratios, 1 - ratio_clip, 1 + ratio_clip) * advantages

        batch_loss = -(T.min(raw_loss, clipped_loss)).mean()
        entropy_loss = -entropy_coeff * entropy

        loss = batch_loss + entropy_loss

        self.run_optimizer(optimizer, loss, actor_parameters)

        new_distributions = model.action_distribution(observations)
        loss = loss.detach()
        entropy = entropy.detach()
        kl = kl_divergence(new_distributions, old_distributions).mean()

        return UpdaterLog(
            loss=loss.item(), divergence=kl.item(), entropy=entropy.item()
        )


class DeterministicPolicyGradient(BaseActorUpdater):
    """
    Deterministic policy gradient used in DDPG: https://spinningup.openai.com/en/latest/algorithms/ddpg.html
    loss = -E[critic(state, actor(state))]

    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        super().__init__(optimizer_class=optimizer_class, max_grad=max_grad)

    def __call__(
        self,
        model: ActorCritic,
        observations: T.Tensor,
        learning_rate: float = 1e-3,
    ) -> UpdaterLog:
        """
        Perform an optimization step

        :param model: an actor critic model
        :param observations:
        :param learning_rate: the learning rate for the optimizer algorithm
        """
        actor_parameters = self._get_model_parameters(model)
        optimizer = self.optimizer_class(actor_parameters, lr=learning_rate)

        actions = model(observations)
        values = model.forward_critics(observations, actions)

        loss = -values.mean()

        self.run_optimizer(optimizer, loss, actor_parameters)

        return UpdaterLog(loss=loss.detach().item())


class SoftPolicyGradient(BaseActorUpdater):
    """
    Policy gradient update used in SAC: https://spinningup.openai.com/en/latest/algorithms/sac.html

    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param squashed_output: whether to squash the actions using tanh, defaults to True
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        squashed_output: bool = True,
        max_grad: float = 0,
    ) -> None:
        super().__init__(optimizer_class=optimizer_class, max_grad=max_grad)
        self.squashed_output = squashed_output

    def __call__(
        self,
        model: ActorCritic,
        observations: T.Tensor,
        learning_rate: float = 1e-3,
        entropy_coeff: float = 0.01,
    ) -> UpdaterLog:
        """
        Perform an optimization step

        :param model: an actor critic model with either 1 or 2 critic networks
        :param observations:
        :param learning_rate: the learning rate for the optimizer algorithm
        :param entropy_coeff: entropy weighting coefficient
        """
        actor_parameters = self._get_model_parameters(model)
        optimizer = self.optimizer_class(actor_parameters, lr=learning_rate)

        distributions = model.action_distribution(observations)
        # use the reparametrization trick for backpropagation
        # https://gregorygundersen.com/blog/2018/04/29/reparameterization/
        actions = distributions.rsample()
        if self.squashed_output:
            actions = T.tanh(actions)
        log_probs = distributions.log_prob(actions)
        entropy = distributions.entropy().mean()

        with T.no_grad():
            if hasattr(model, "critic2"):
                values1, values2 = model.forward_critics(observations, actions)
                values = T.min(values1, values2)
            else:
                values = model.forward_critics(observations, actions)

        loss = (entropy_coeff * log_probs - values).mean()

        self.run_optimizer(optimizer, loss, actor_parameters)

        return UpdaterLog(loss=loss.detach().item(), entropy=entropy.detach().item())



from typing import Optional, Tuple, Union

import torch as T

from pearll.common.type_aliases import Tensor
from pearll.common.utils import to_torch


def trainable_parameters(model: T.nn.Module) -> list:
    return [p for p in model.parameters() if p.requires_grad]


def prod(val):
    res = 1
    for ele in val:
        res *= ele
    return res


def get_mlp_size(data_shape: Union[int, Tuple[int]]) -> int:
    """
    Convert the shape of some data into an integer size that defines
    how many neurons to create in an MLP layer

    :param data_shape: the tuple shape or already converted integer size
    :return: the integer size
    """

    if isinstance(data_shape, tuple):
        data_shape = prod(list(data_shape))
    return data_shape


def concat_obs_actions(observations: Tensor, actions: Optional[Tensor]) -> T.Tensor:
    if actions is not None:
        observations, actions = to_torch(observations, actions)
        if observations.dim() == 0:
            observations = observations.unsqueeze(0)
        if actions.dim() == 0:
            actions = actions.unsqueeze(0)
        return T.cat([observations, actions], dim=-1).float()
    return to_torch(observations).float()



from typing import List, Optional, Type

import numpy as np
import torch as T
from gym import Env

from pearll.agents.base_agents import BaseAgent
from pearll.buffers.base_buffer import BaseBuffer
from pearll.buffers.replay_buffer import ReplayBuffer
from pearll.callbacks.base_callback import BaseCallback
from pearll.common.type_aliases import Log
from pearll.common.utils import get_space_shape
from pearll.explorers.base_explorer import BaseExplorer
from pearll.models.actor_critics import ActorCritic, Critic, EpsilonGreedyActor
from pearll.models.encoders import IdentityEncoder
from pearll.models.heads import DiscreteQHead
from pearll.models.torsos import MLP
from pearll.models.utils import get_mlp_size
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    MiscellaneousSettings,
    OptimizerSettings,
    Settings,
)
from pearll.signal_processing.return_estimators import TD_zero
from pearll.updaters.critics import BaseCriticUpdater, DiscreteQRegression


def get_default_model(env: Env):
    observation_shape = get_space_shape(env.observation_space)
    action_size = env.action_space.n
    observation_size = get_mlp_size(observation_shape)

    encoder = IdentityEncoder()
    torso = MLP(layer_sizes=[observation_size, 64, 32], activation_fn=T.nn.ReLU)
    head = DiscreteQHead(input_shape=32, output_shape=action_size)

    actor = EpsilonGreedyActor(
        critic_encoder=encoder, critic_torso=torso, critic_head=head
    )
    critic = Critic(encoder=encoder, torso=torso, head=head, create_target=True)

    return ActorCritic(actor, critic)


class DQN(BaseAgent):
    """
    DQN Algorithm

    :param env: the gym-like environment to be used
    :param model: the neural network model
    :param td_gamma: trajectory discount factor
    :param updater_class: critic updater class
    :param optimizer_settings: optimizer settings
    :param buffer_class: the buffer class for storing and sampling trajectories
    :param buffer_settings: settings for the buffer
    :param action_explorer_class: the explorer class for random search at beginning of training and
        adding noise to actions
    :param explorer_settings: settings for the action explorer
    :param callbacks: an optional list of callbacks (e.g. if you want to save the model)
    :param callback_settings: settings for callbacks
    :param logger_settings: settings for the logger
    :param misc_settings: settings for miscellaneous parameters
    """

    def __init__(
        self,
        env: Env,
        model: ActorCritic,
        td_gamma: float = 0.99,
        updater_class: Type[BaseCriticUpdater] = DiscreteQRegression,
        optimizer_settings: OptimizerSettings = OptimizerSettings(),
        buffer_class: Type[BaseBuffer] = ReplayBuffer,
        buffer_settings: BufferSettings = BufferSettings(),
        action_explorer_class: Type[BaseExplorer] = BaseExplorer,
        explorer_settings: ExplorerSettings = ExplorerSettings(start_steps=1000),
        callbacks: Optional[List[Type[BaseCallback]]] = None,
        callback_settings: Optional[List[Settings]] = None,
        logger_settings: LoggerSettings = LoggerSettings(),
        misc_settings: MiscellaneousSettings = MiscellaneousSettings(),
    ) -> None:
        model = model or get_default_model(env)
        super().__init__(
            env,
            model,
            action_explorer_class=action_explorer_class,
            explorer_settings=explorer_settings,
            buffer_class=buffer_class,
            buffer_settings=buffer_settings,
            logger_settings=logger_settings,
            callbacks=callbacks,
            callback_settings=callback_settings,
            misc_settings=misc_settings,
        )
        self.updater = updater_class(
            loss_class=optimizer_settings.loss_class,
            optimizer_class=optimizer_settings.optimizer_class,
            max_grad=optimizer_settings.max_grad,
        )

        self.learning_rate = optimizer_settings.learning_rate
        self.td_gamma = td_gamma

    def _fit(
        self, batch_size: int, actor_epochs: int = 1, critic_epochs: int = 1
    ) -> Log:
        critic_losses = np.zeros(shape=(critic_epochs))
        for i in range(critic_epochs):
            trajectories = self.buffer.sample(batch_size=batch_size, flatten_env=False)

            with T.no_grad():
                next_q_values = self.model.forward_target_critics(
                    trajectories.next_observations
                )
                next_q_values = T.unsqueeze(next_q_values.max(dim=-1)[0], dim=-1)
                target_q_values = TD_zero(
                    trajectories.rewards,
                    next_q_values,
                    trajectories.dones,
                    self.td_gamma,
                )

            updater_log = self.updater(
                self.model,
                trajectories.observations,
                target_q_values,
                trajectories.actions,
                learning_rate=self.learning_rate,
            )
            critic_losses[i] = updater_log.loss

        self.model.assign_targets()

        return Log(critic_loss=critic_losses.mean())



from enum import Enum


class NetworkType(Enum):
    MLP = "mlp"
    PARAMETER = "parameter"


class TrajectoryType(Enum):
    NUMPY = "numpy"
    TORCH = "torch"


class FrequencyType(Enum):
    EPISODE = "episode"
    STEP = "step"


class GoalSelectionStrategy(Enum):
    """Goal selection strategy for HER Replay Buffer"""

    FINAL = "final"
    FUTURE = "future"
    EPISODE = "episode"


class Distribution(Enum):
    """Distribution types"""

    NORMAL = "normal"
    UNIFORM = "uniform"



"""
Methods for mutating a population
https://www.tutorialspoint.com/genetic_algorithms/genetic_algorithms_mutation.htm
"""

import numpy as np
from gym import Space
from gym.spaces.discrete import Discrete
from gym.spaces.multi_discrete import MultiDiscrete

from pearll.common.utils import get_space_range


def _sample_indices(
    population: np.ndarray,
    mutation_rate: float = 0.1,
) -> np.ndarray:
    """
    Samples mutation indices from a population.

    :param population: the population of individuals to mutate
    :param mutation_rate: the probability of mutating an individual
    :return: the population indices to mutate
    """
    # Get indices of individuals to mutate
    mutation_indices = np.random.choice(
        np.arange(population.shape[0]),
        size=int(population.shape[0] * mutation_rate),
        replace=False,
    )

    return mutation_indices


def gaussian_mutation(
    population: np.ndarray,
    action_space: Space,
    mutation_rate: float = 0.1,
    mutation_std: float = 0.5,
) -> np.ndarray:
    """
    Mutates a population using a Gaussian mutation operator.

    :param population: the population of individuals to mutate
    :param action_space: the action space of the environment
    :param mutation_rate: the probability of mutating an individual
    :param mutation_std: the standard deviation of the Gaussian distribution used to mutate an individual
    :return: the mutated population
    """
    space_range = get_space_range(action_space)
    # Get indices of individuals to mutate
    mutation_indices = _sample_indices(population, mutation_rate)

    # Mutate individuals
    new_population = population.copy().astype(np.float32)
    for i in mutation_indices:
        new_population[i] += np.random.normal(0, mutation_std, population.shape[-1])

    # Discretize population as required
    if isinstance(action_space, (Discrete, MultiDiscrete)):
        new_population = np.round(new_population).astype(np.int32)

    return np.clip(new_population, space_range[0], space_range[1])


def uniform_mutation(
    population: np.ndarray,
    action_space: Space,
    mutation_rate: float = 0.1,
) -> np.ndarray:
    """
    Mutates a population using a uniform mutation operator.

    :param population: the population of individuals to mutate
    :param action_space: the action space of the environment
    :param mutation_rate: the probability of mutating an individual
    :return: the mutated population
    """
    space_range = get_space_range(action_space)
    # Get indices of individuals to mutate
    mutation_indices = _sample_indices(population, mutation_rate)

    # Mutate individuals
    new_population = population.copy().astype(np.float32)
    for i in mutation_indices:
        new_population[i] += np.random.uniform(-1, 1, population.shape[1])

    # Discretize population as required
    if isinstance(action_space, (Discrete, MultiDiscrete)):
        new_population = np.round(new_population).astype(np.int32)

    return np.clip(new_population, space_range[0], space_range[1])



from typing import List, Optional, Type

from gym import Env

from pearll.agents.base_agents import BaseAgent
from pearll.buffers.base_buffer import BaseBuffer
from pearll.callbacks.base_callback import BaseCallback
from pearll.common.type_aliases import Log
from pearll.explorers.base_explorer import BaseExplorer
from pearll.models.actor_critics import ActorCritic
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    MiscellaneousSettings,
    Settings,
)


class YourRLAgent(BaseAgent):
    """
    A template for a new agent :)

    Add here any other modules you like; for example, an actor updater or a critic updater. If you
    want to use a new module which inherits from an existing module and need to pass through unique
    settings, simply inherit the setting object from which the module is based and update as required
    before passing here.

    For example:
    ```
    @dataclass
    class HERBufferSettings(BufferSettings):
        goal_selection_strategy: Union[str, GoalSelectionStrategy] = "future"
        n_sampled_goal: int = 4
    ```

    :param env: the gym-like environment to be used
    :param model: the neural network model
    :param buffer_class: the buffer class for storing and sampling trajectories
    :param buffer_settings: settings for the buffer
    :param action_explorer_class: the explorer class for random search at beginning of training and
        adding noise to actions
    :param explorer settings: settings for the action explorer
    :param callbacks: an optional list of callbacks (e.g. if you want to save the model)
    :param callback_settings: settings for callbacks
    :param logger_settings: settings for the logger
    :param misc_settings: settings for miscellaneous parameters
    """

    def __init__(
        self,
        env: Env,
        model: ActorCritic,
        buffer_class: Type[BaseBuffer] = ...,
        buffer_settings: BufferSettings = ...,
        action_explorer_class: Type[BaseExplorer] = ...,
        explorer_settings: ExplorerSettings = ...,
        callbacks: Optional[List[Type[BaseCallback]]] = None,
        callback_settings: Optional[List[Settings]] = None,
        logger_settings: LoggerSettings = ...,
        misc_settings: MiscellaneousSettings = ...,
    ) -> None:
        super().__init__(
            env,
            model,
            action_explorer_class=action_explorer_class,
            explorer_settings=explorer_settings,
            buffer_class=buffer_class,
            buffer_settings=buffer_settings,
            logger_settings=logger_settings,
            callbacks=callbacks,
            callback_settings=callback_settings,
            misc_settings=misc_settings,
        )

    def _fit(
        self, batch_size: int, actor_epochs: int = 1, critic_epochs: int = 1
    ) -> Log:
        """
        Specify your algorithm logic here! Should call your critic_updater and actor_updater
        and return a `Log` object, details of which can be found in pearll/common/type_aliases.py
        """



from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple, Type, Union

import torch as T

from pearll.common.enumerations import NetworkType
from pearll.models.torsos import MLP
from pearll.models.utils import get_mlp_size

################################### BASE CLASSES ###################################


class BaseCriticHead(T.nn.Module):
    """
    The base class for the critic head network

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param output_shape: the output shape of the network, can be the tuple shape or simplified integer output size
    :param network_type: the type of network used
    :param activation_fn: the activation function after each layer
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        output_shape: Union[int, Tuple[int]],
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = None,
    ):
        super().__init__()
        network_type = NetworkType(network_type.lower())
        if network_type == NetworkType.MLP:
            input_size = get_mlp_size(input_shape)
            output_size = get_mlp_size(output_shape)
            self.model = MLP(
                layer_sizes=[input_size, output_size], activation_fn=activation_fn
            )
        else:
            raise NotImplementedError(f"{network_type} hasn't been implemented yet.")

    def forward(self, input: T.Tensor) -> T.Tensor:
        return self.model(input)


class BaseActorHead(T.nn.Module, ABC):
    """
    The base class for the actor head network

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param output_shape: the output shape of the network, can be the tuple shape or simplified integer output size
    :param network_type: the type of network used
    :param activation_fn: the activation function after each layer
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        output_shape: Union[int, Tuple[int]],
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = None,
    ):
        super().__init__()
        network_type = NetworkType(network_type.lower())
        if network_type == NetworkType.MLP:
            input_size = get_mlp_size(input_shape)
            output_size = get_mlp_size(output_shape)
            self.model = MLP(
                layer_sizes=[input_size, output_size], activation_fn=activation_fn
            )
        else:
            raise NotImplementedError(f"{network_type} hasn't been implemented yet.")

    @abstractmethod
    def action_distribution(
        self, input: T.Tensor
    ) -> Optional[T.distributions.Distribution]:
        """
        Get the action distribution given a latent input from the torso

        :param input: torso network output
        :return: the distribution of the actor, returns None if deterministic
        """

    def forward(self, input: T.Tensor) -> T.Tensor:
        distribution = self.action_distribution(input)
        if distribution is None:
            return self.model(input)
        return distribution.sample()


class BaseEnvHead(T.nn.Module):
    """
    The base class for the environment head network

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param output_shape: the output shape of the network, can be the tuple shape or simplified integer output size
    :param network_type: the type of network used
    :param activation_fn: the activation function after each layer
    :param dtype: optional conversion of network output to dtype
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        output_shape: Union[int, Tuple[int]],
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = None,
        dtype: Optional[str] = None,
    ):
        super().__init__()
        self.type_map = {
            "bool": "torch.BoolTensor",
            "int": "torch.IntTensor",
            "float": "torch.FloatTensor",
            "long": "torch.LongTensor",
            "double": "torch.DoubleTensor",
        }
        self.dtype = self.type_map[dtype] if dtype is not None else None

        network_type = NetworkType(network_type.lower())
        if network_type == NetworkType.MLP:
            input_size = get_mlp_size(input_shape)
            output_size = get_mlp_size(output_shape)
            self.model = MLP(
                layer_sizes=[input_size, output_size], activation_fn=activation_fn
            )
        else:
            raise NotImplementedError(f"{network_type} hasn't been implemented yet.")

    def forward(self, input: T.Tensor) -> T.Tensor:
        return self.model(input)


class DummyHead(T.nn.Module):
    def __init__(self):
        super().__init__()

    def action_distribution(
        self, input: T.Tensor
    ) -> Optional[T.distributions.Distribution]:
        return None


################################### CRITIC HEADS ###################################


class ValueHead(BaseCriticHead):
    """
    Use this head if you want to model the value function or the continuous Q function.

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param activation_fn: the activation function after each layer
    :param network_type: the type of network used
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = None,
    ):
        super().__init__(
            input_shape=input_shape,
            output_shape=1,
            network_type=network_type,
            activation_fn=activation_fn,
        )


# The continuous Q function has the same structure as
# the value function so we can use the same object.
ContinuousQHead = ValueHead


class DiscreteQHead(BaseCriticHead):
    """
    Use this head if you want to model the discrete Q function.

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param output_shape: the output shape of the network, can be the tuple shape or simplified integer output size
    :param activation_fn: the activation function after each layer
    :param network_type: the type of network used
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        output_shape: Union[int, Tuple[int]],
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = None,
    ):
        super().__init__(
            input_shape=input_shape,
            output_shape=output_shape,
            network_type=network_type,
            activation_fn=activation_fn,
        )


################################### ACTOR HEADS ###################################


class DeterministicHead(BaseActorHead):
    """
    Use this head if you want a deterministic actor.

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param action_shape: the output shape of the network, can be the tuple shape or simplified integer output size
    :param network_type: the type of network used
    :param activation_fn: the activation function after each layer
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        action_shape: Union[int, Tuple[int]],
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = None,
    ):
        super().__init__(
            input_shape=input_shape,
            output_shape=action_shape,
            network_type=network_type,
            activation_fn=activation_fn,
        )

    def action_distribution(
        self, input: T.Tensor
    ) -> Optional[T.distributions.Distribution]:
        return None


class CategoricalHead(BaseActorHead):
    """
    Use this head for a categorical actor.

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param action_size: the dimension of the action vector
    :param network_type: the type of network used
    :param activation_fn: the activation function after each layer
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        action_size: int,
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = None,
    ):
        super().__init__(
            input_shape=input_shape,
            output_shape=action_size,
            network_type=network_type,
            activation_fn=activation_fn,
        )

    def action_distribution(
        self, input: T.Tensor
    ) -> Optional[T.distributions.Distribution]:
        return T.distributions.Categorical(logits=self.model(input))


class DiagGaussianHead(BaseActorHead):
    """
    Use this head if you want a policy obeying a diagonal gaussian distribution.

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param action_size: the dimension of the action vector
    :param mean_network_type: network type of the network calculating the mean
    :param mean_activation: activation function of the output of the mean network
    :param log_std_network_type: network type of the network calculating the log std
    :param log_std_activation: activation function of the output of the log std network
    :param log_std_init: if a parameter is used for log std, what initial value should it have
    :param min_log_std: minimum log std of the action distribution
    :param max_log_std: maximum log std of the action distribution
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        action_size: int,
        mean_network_type: str = "mlp",
        mean_activation: Type[T.nn.Module] = T.nn.Tanh,
        log_std_network_type: str = "parameter",
        log_std_activation: Type[T.nn.Module] = T.nn.Softplus,
        log_std_init: float = 0.0,
        min_log_std: Optional[float] = None,
        max_log_std: Optional[float] = None,
    ):
        super().__init__(
            input_shape=input_shape,
            output_shape=action_size,
            network_type=mean_network_type,
            activation_fn=mean_activation,
        )
        self.mean_network = self.model

        self.min_log_std = min_log_std
        self.max_log_std = max_log_std
        log_std_network_type = NetworkType(log_std_network_type.lower())

        if log_std_network_type == NetworkType.MLP:
            input_size = get_mlp_size(input_shape)
            self.log_std_network = MLP(
                layer_sizes=[input_size, action_size], activation_fn=log_std_activation
            )
        elif log_std_network_type == NetworkType.PARAMETER:
            self.log_std_network = T.nn.Parameter(
                T.ones(action_size) * log_std_init, requires_grad=True
            )
        else:
            raise NotImplementedError(
                f"{log_std_network_type} hasn't been implemented for the log std network yet."
            )

    def action_distribution(
        self, input: T.Tensor
    ) -> Optional[T.distributions.Distribution]:
        mean = self.mean_network(input)
        if isinstance(self.log_std_network, T.nn.Parameter):
            log_std = self.log_std_network
        else:
            log_std = self.log_std_network(input)
        if self.min_log_std or self.max_log_std:
            log_std = T.clamp(log_std, self.min_log_std, self.max_log_std)
        return T.distributions.Normal(mean, log_std.exp())


################################### ENVIRONMENT HEADS ###################################


class BoxHead(BaseEnvHead):
    """
    Use this head to model a Box gym environment.

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param space_shape: the output shape of the network, can be the tuple shape or simplified integer output size
    :param network_type: the type of network used
    :param activation_fn: the activation function after each layer
    :param dtype: optional conversion of network output to dtype
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        space_shape: Union[int, Tuple[int]],
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = None,
        dtype: Optional[dict] = None,
    ) -> None:
        super().__init__(
            input_shape=input_shape,
            output_shape=space_shape,
            network_type=network_type,
            activation_fn=activation_fn,
            dtype=dtype,
        )

    def forward(self, input: T.Tensor) -> Any:
        out = self.model(input)
        if self.dtype is not None:
            out = out.type(self.dtype)
        if out.squeeze().dim() == 0:
            return out.squeeze().item()
        return out.squeeze()


class DiscreteHead(BaseEnvHead):
    """
    Use this head to model a Discrete gym environment.

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param space_size: the output size of the network
    :param network_type: the type of network used
    :param activation_fn: the activation function after each layer
    :param dtype: optional conversion of network output to dtype
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        space_size: int,
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = T.nn.Softmax,
        dtype: Optional[dict] = None,
    ) -> None:
        super().__init__(
            input_shape=input_shape,
            output_shape=space_size,
            network_type=network_type,
            activation_fn=activation_fn,
            dtype=dtype,
        )

    def forward(self, input: T.Tensor) -> Any:
        out = T.argmax(self.model(input), dim=-1)
        if self.dtype is not None:
            out = out.type(self.dtype)
        if out.squeeze().dim() == 0:
            return out.squeeze().item()
        return out.squeeze()


class MultiDiscreteHead(BaseEnvHead):
    """
    Use this head to model a MultiDiscrete gym environment.

    :param input_shape: the input shape to the head network, can be the tuple shape or simplified integer input size
    :param space_shape: the output shape of the network, can be the tuple shape or simplified integer output size
    :param network_type: the type of network used
    :param activation_fn: the activation function after each layer
    :param output_map: optional mapping of the network output to a different output space
    :param dtype: optional conversion of network output to dtype
    """

    def __init__(
        self,
        input_shape: Union[int, Tuple[int]],
        space_shape: Union[int, Tuple[int]],
        network_type: str = "mlp",
        activation_fn: Optional[Type[T.nn.Module]] = None,
        output_map: Optional[dict] = None,
        dtype: Optional[dict] = None,
    ):
        super().__init__(
            input_shape=input_shape,
            output_shape=space_shape,
            network_type=network_type,
            activation_fn=activation_fn,
            dtype=dtype,
        )
        self.output_map = output_map

    def forward(self, input: T.Tensor) -> Any:
        out = T.argmax(self.model(input), dim=-1, keepdim=True)
        if self.dtype is not None:
            out = out.type(self.dtype)
        if self.output_map is not None:
            out = T.Tensor([self.output_map[x.item()] for x in out])
        if out.squeeze().dim() == 0:
            return out.squeeze().item()
        return out.squeeze()



from typing import List, Optional, Tuple, Type

import numpy as np
import torch as T
from gym import Env

from pearll.agents import BaseAgent
from pearll.buffers import BaseBuffer, ReplayBuffer
from pearll.callbacks.base_callback import BaseCallback
from pearll.common.enumerations import FrequencyType
from pearll.common.type_aliases import Log, Observation, Trajectories
from pearll.explorers.base_explorer import BaseExplorer
from pearll.models import ModelEnv
from pearll.models.actor_critics import ActorCritic
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    MiscellaneousSettings,
    OptimizerSettings,
    Settings,
)
from pearll.signal_processing import return_estimators
from pearll.updaters.critics import BaseCriticUpdater, DiscreteQRegression
from pearll.updaters.environment import BaseDeepUpdater, DeepRegression


class DynaQ(BaseAgent):
    """
    Dyna-Q model based RL algorithm.

    :param env: the gym-like environment to be used
    :param agent_model: the agent model to be used
    :param env_model: the environment model to be used
    :param td_gamma: trajectory discount factor
    :param agent_updater_class: the updater class for the agent critic
    :param agent_optimizer_settings: the settings for the agent updater
    :param obs_updater_class: the updater class for the observation function in the environment model
    :param obs_optimizer_settings: the settings for the observation updater
    :param reward_updater_class: the updater class for the reward function in the environment model
    :param reward_optimizer_settings: the settings for the reward updater
    :param done_updater_class: the updater class for the done function in the environment model
    :param done_optimizer_settings: the settings for the done updater
    :param buffer_class: the buffer class for storing and sampling trajectories
    :param buffer_settings: settings for the buffer
    :param action_explorer_class: the explorer class for random search at beginning of training and
        adding noise to actions
    :param explorer settings: settings for the action explorer
    :param callbacks: an optional list of callbacks (e.g. if you want to save the model)
    :param callback_settings: settings for callbacks
    :param logger_settings: settings for the logger
    :param misc_settings: settings for miscellaneous parameters
    """

    def __init__(
        self,
        env: Env,
        agent_model: ActorCritic,
        env_model: ModelEnv,
        td_gamma: float = 0.99,
        agent_updater_class: Type[BaseCriticUpdater] = DiscreteQRegression,
        agent_optimizer_settings: OptimizerSettings = OptimizerSettings(),
        obs_updater_class: BaseDeepUpdater = DeepRegression,
        obs_optimizer_settings: OptimizerSettings = OptimizerSettings(),
        reward_updater_class: Type[BaseDeepUpdater] = DeepRegression,
        reward_optimizer_settings: OptimizerSettings = OptimizerSettings(),
        done_updater_class: Optional[Type[BaseDeepUpdater]] = None,
        done_optimizer_settings: OptimizerSettings = OptimizerSettings(
            loss_class=T.nn.BCELoss()
        ),
        buffer_class: Type[BaseBuffer] = ReplayBuffer,
        buffer_settings: BufferSettings = BufferSettings(),
        action_explorer_class: Type[BaseExplorer] = BaseExplorer,
        explorer_settings: ExplorerSettings = ExplorerSettings(start_steps=100),
        callbacks: Optional[List[Type[BaseCallback]]] = None,
        callback_settings: Optional[List[Settings]] = None,
        logger_settings: LoggerSettings = LoggerSettings(),
        misc_settings: MiscellaneousSettings = MiscellaneousSettings(),
    ) -> None:
        super().__init__(
            env=env,
            model=agent_model,
            action_explorer_class=action_explorer_class,
            explorer_settings=explorer_settings,
            buffer_class=buffer_class,
            buffer_settings=buffer_settings,
            logger_settings=logger_settings,
            callbacks=callbacks,
            callback_settings=callback_settings,
            misc_settings=misc_settings,
        )
        self.env_model = env_model
        self.td_gamma = td_gamma
        self.learning_rate = agent_optimizer_settings.learning_rate
        self.policy_updater = agent_updater_class(
            loss_class=agent_optimizer_settings.loss_class,
            optimizer_class=agent_optimizer_settings.optimizer_class,
            max_grad=agent_optimizer_settings.max_grad,
        )
        self.obs_updater = obs_updater_class(
            loss_class=obs_optimizer_settings.loss_class,
            optimizer_class=obs_optimizer_settings.optimizer_class,
            max_grad=obs_optimizer_settings.max_grad,
        )
        self.reward_updater = reward_updater_class(
            loss_class=reward_optimizer_settings.loss_class,
            optimizer_class=reward_optimizer_settings.optimizer_class,
            max_grad=reward_optimizer_settings.max_grad,
        )
        if self.env_model.done_fn is not None:
            self.done_updater = done_updater_class(
                loss_class=done_optimizer_settings.loss_class,
                optimizer_class=done_optimizer_settings.optimizer_class,
                max_grad=done_optimizer_settings.max_grad,
            )

        self.model_step = 0
        self.model_episode = 0

    def step_model_env(
        self, observation: Observation, num_steps: int = 1
    ) -> np.ndarray:
        """
        Step the agent in the model environment

        :param observation: the starting observation to step from
        :param num_steps: how many steps to take
        :return: the final observation after all steps have been done
        """
        self.model.eval()
        for _ in range(num_steps):
            action = self.action_explorer(self.model, observation, self.step)
            next_observation, reward, done, _ = self.env_model.step(observation, action)
            self.buffer.add_trajectory(
                observation, action, reward, next_observation, done
            )
            self.logger.debug(
                Trajectories(observation, action, reward, next_observation, done)
            )
            observation = next_observation

            if done:
                observation = self.env_model.reset()
                self.model_episode += 1
            self.model_step += 1

        return observation

    def _fit_model_env(self, batch_size: int, epochs: int = 1) -> None:
        """
        Fit the model environment

        :param batch_size: the batch size to use
        :param epochs: how many epochs to fit for
        """
        obs_loss = np.zeros(epochs)
        reward_loss = np.zeros(epochs)
        done_loss = np.zeros(epochs)
        for i in range(epochs):
            trajectories = self.buffer.sample(batch_size)
            obs_update_log = self.obs_updater(
                model=self.env_model.observation_fn,
                observations=trajectories.observations,
                actions=trajectories.actions,
                targets=trajectories.next_observations,
                learning_rate=self.learning_rate,
            )
            obs_loss[i] = obs_update_log.loss
            reward_update_log = self.reward_updater(
                model=self.env_model.reward_fn,
                observations=trajectories.observations,
                actions=trajectories.actions,
                targets=trajectories.rewards,
                learning_rate=self.learning_rate,
            )
            reward_loss[i] = reward_update_log.loss
            if self.env_model.done_fn is not None:
                done_update_log = self.done_updater(
                    model=self.env_model.done_fn,
                    observations=trajectories.observations,
                    actions=trajectories.actions,
                    targets=trajectories.dones,
                    learning_rate=self.learning_rate,
                )
                done_loss[i] = done_update_log.loss

        self.logger.debug(f"obs_loss: {obs_loss.mean()}")
        self.logger.debug(f"reward_loss: {reward_loss.mean()}")
        if self.env_model.done_fn is not None:
            self.logger.debug(f"done_loss: {done_loss.mean()}")

    def _fit(
        self, batch_size: int, actor_epochs: int = 1, critic_epochs: int = 1
    ) -> Log:
        critic_losses = np.zeros(shape=(critic_epochs))
        for i in range(critic_epochs):
            trajectories = self.buffer.sample(batch_size=batch_size, flatten_env=False)

            with T.no_grad():
                next_q_values = self.model.forward_target_critics(
                    trajectories.next_observations
                )
                next_q_values = T.unsqueeze(next_q_values.max(dim=-1)[0], dim=-1)
                target_q_values = return_estimators.TD_zero(
                    trajectories.rewards,
                    next_q_values,
                    trajectories.dones,
                    self.td_gamma,
                )

            updater_log = self.policy_updater(
                self.model,
                trajectories.observations,
                target_q_values,
                trajectories.actions,
                learning_rate=self.learning_rate,
            )
            critic_losses[i] = updater_log.loss

        self.model.assign_targets()

        return Log(critic_loss=np.mean(critic_losses))

    def fit(
        self,
        env_steps: int,
        plan_steps: int,
        env_batch_size: int,
        plan_batch_size: int,
        actor_epochs: int = 1,
        critic_epochs: int = 1,
        env_epochs: int = 1,
        env_train_frequency: Tuple[str, int] = ("step", 1),
        plan_train_frequency: Tuple[str, int] = ("step", 1),
        no_model_steps: int = 0,
    ) -> None:
        """
        Train the agent in the environment

        1. Collect samples in the real environment.
        2. Train the model environment on samples collected.
        3. Collect samples in the model environment.
        4. Train the agent on samples collected in both environments.

        :param env_steps: total number of real environment steps to train over
        :param plan_steps: number of model environment steps to run each planning phase
        :param env_batch_size: minibatch size for the model environment to make a single gradient descent step on
        :param plan_batch_size: minibatch size for the agent to make a single gradient descent step on
        :param actor_epochs: how many times to update the actor network in each training step
        :param critic_epochs: how many times to update the critic network in each training step
        :param env_epochs: how many times to update the model environment in each training step
        :param env_train_frequency: the number of steps or episodes to run in the real environment before running a model environment training step.
            To run every n episodes, use `("episode", n)`.
            To run every n steps, use `("step", n)`.
        :param plan_train_frequency: the number of steps or episodes to run in the model environment before running an agent training step.
            To run every n episodes, use `("episode", n)`.
            To run every n steps, use `("step", n)`.
        :param no_model_steps: number of steps to run without collecting trajectories from the model environment.
        """
        env_train_frequency = (
            FrequencyType(env_train_frequency[0].lower()),
            env_train_frequency[1],
        )
        plan_train_frequency = (
            FrequencyType(plan_train_frequency[0].lower()),
            plan_train_frequency[1],
        )

        # We can pre-calculate how many training steps to run if train frequency is in steps rather than episodes
        if env_train_frequency[0] == FrequencyType.STEP:
            env_steps = env_steps // env_train_frequency[1]
        if plan_train_frequency[0] == FrequencyType.STEP:
            plan_steps = plan_steps // plan_train_frequency[1]

        observation = self.env.reset()
        for _ in range(env_steps):
            self.logger.debug("REAL ENVIRONMENT")
            # Step for number of steps specified
            if env_train_frequency[0] == FrequencyType.STEP:
                observation = self.step_env(
                    observation=observation, num_steps=env_train_frequency[1]
                )
            # Step for number of episodes specified
            elif env_train_frequency[0] == FrequencyType.EPISODE:
                start_episode = self.episode
                end_episode = start_episode + env_train_frequency[1]
                while self.episode != end_episode:
                    observation = self.step_env(observation=observation)
                if self.step >= env_steps:
                    break

            if self.done:
                break

            # Update the environment model
            self._fit_model_env(batch_size=env_batch_size, epochs=env_epochs)

            if self.step < no_model_steps:
                # Update the agent model
                self.model.train()
                train_log = self._fit(
                    batch_size=plan_batch_size,
                    actor_epochs=actor_epochs,
                    critic_epochs=critic_epochs,
                )
                self.model.update_global()
                self.logger.add_train_log(train_log)
            else:
                self.logger.debug("MODEL ENVIRONMENT")
                # Plan for number of steps specified
                model_obs = self.env_model.reset()
                for _ in range(plan_steps):
                    # Step for number of steps specified
                    if plan_train_frequency[0] == FrequencyType.STEP:
                        model_obs = self.step_model_env(
                            observation=model_obs, num_steps=plan_train_frequency[1]
                        )
                    # Step for number of episodes specified
                    elif plan_train_frequency[0] == FrequencyType.EPISODE:
                        start_episode = self.model_episode
                        end_episode = start_episode + plan_train_frequency[1]
                        while self.model_episode != end_episode:
                            observation = self.step_model_env(observation=observation)
                        if self.model_step >= plan_steps:
                            break

                    # Update the agent model
                    self.model.train()
                    train_log = self._fit(
                        batch_size=plan_batch_size,
                        actor_epochs=actor_epochs,
                        critic_epochs=critic_epochs,
                    )
                    self.model.update_global()
                    self.logger.add_train_log(train_log)

                self.buffer.reset()
